from __future__ import annotations

from pydiverse.transform import C
from pydiverse.transform.core.verbs import (
    arrange,
    filter,
    group_by,
    mutate,
    select,
    summarise,
)
from pydiverse.transform.errors import ExpressionTypeError, FunctionTypeError
from tests.util import assert_result_equal


def test_ungrouped(df3):
    assert_result_equal(df3, lambda t: t >> summarise(mean3=t.col3.mean()))
    assert_result_equal(
        df3,
        lambda t: t >> summarise(mean3=t.col3.mean(), mean4=t.col4.mean()),
    )


def test_simple_grouped(df3):
    assert_result_equal(
        df3,
        lambda t: t >> group_by(t.col1) >> summarise(mean3=t.col3.mean()),
    )


def test_multi_grouped(df3):
    assert_result_equal(
        df3,
        lambda t: t >> group_by(t.col1, t.col2) >> summarise(mean3=t.col3.mean()),
    )


def test_chained_summarised(df3):
    assert_result_equal(
        df3,
        lambda t: t
        >> group_by(t.col1, t.col2)
        >> summarise(mean3=t.col3.mean())
        >> summarise(mean_of_mean3=C.mean3.mean()),
    )

    assert_result_equal(
        df3,
        lambda t: t
        >> mutate(k=(C.col1 + C.col2) * C.col4)
        >> group_by(C.k)
        >> summarise(x=C.col4.mean())
        >> summarise(y=C.k.mean()),
    )


def test_nested(df3):
    assert_result_equal(
        df3,
        lambda t: t
        >> group_by(t.col1, t.col2)
        >> summarise(mean_of_mean3=t.col3.mean().mean()),
        exception=FunctionTypeError,
    )


def test_select(df3):
    assert_result_equal(
        df3,
        lambda t: t
        >> group_by(t.col1, t.col2)
        >> summarise(mean3=t.col3.mean())
        >> select(t.col1, C.mean3, t.col2),
    )


def test_mutate(df3):
    assert_result_equal(
        df3,
        lambda t: t
        >> group_by(t.col1, t.col2)
        >> summarise(mean3=t.col3.mean())
        >> mutate(x10=C.mean3 * 10),
    )


def test_filter(df3):
    assert_result_equal(
        df3,
        lambda t: t
        >> group_by(t.col1, t.col2)
        >> summarise(mean3=t.col3.mean())
        >> filter(C.mean3 <= 2.0),
    )


# def test_filter_argument(df3):
#     assert_result_equal(
#         df3,
#         lambda t: t
#         >> group_by(t.col2)
#         >> summarise(u=t.col4.sum(filter=(t.col1 != 0))),
#     )

#     assert_result_equal(
#         df3,
#         lambda t: t
#         >> group_by(t.col4, t.col1)
#         >> summarise(
#             u=(t.col3 * t.col4 - t.col2).sum(
#                 filter=(t.col5.isin("a", "e", "i", "o", "u"))
#             )
#         ),
#     )


def test_arrange(df3):
    assert_result_equal(
        df3,
        lambda t: t
        >> group_by(t.col1, t.col2)
        >> summarise(mean3=t.col3.mean())
        >> arrange(C.mean3),
    )

    assert_result_equal(
        df3,
        lambda t: t
        >> arrange(-t.col4)
        >> group_by(t.col1, t.col2)
        >> summarise(mean3=t.col3.mean())
        >> arrange(C.mean3),
    )


def test_intermediate_select(df3):
    # Check that subqueries happen transparently
    assert_result_equal(
        df3,
        lambda t: t
        >> group_by(t.col1, t.col2)
        >> summarise(x=t.col4.mean())
        >> mutate(x2=C.x * 2)
        >> select()
        >> summarise(y=(C.x - C.x2).min()),
    )


def test_not_summarising(df4):
    assert_result_equal(
        df4, lambda t: t >> summarise(x=C.col1), exception=FunctionTypeError
    )


def test_none(df4):
    assert_result_equal(
        df4, lambda t: t >> summarise(x=None), exception=ExpressionTypeError
    )


# TODO: Implement more test cases for summarise verb


# Test specific operations


def test_op_min(df4):
    assert_result_equal(
        df4,
        lambda t: t
        >> group_by(t.col1)
        >> summarise(**{c._.name + "_min": c.min() for c in t}),
    )


def test_op_max(df4):
    assert_result_equal(
        df4,
        lambda t: t
        >> group_by(t.col1)
        >> summarise(**{c._.name + "_max": c.max() for c in t}),
    )


def test_op_any(df4):
    assert_result_equal(
        df4,
        lambda t: t >> group_by(t.col1) >> summarise(any=(C.col1 == C.col2).any()),
    )
    assert_result_equal(
        df4,
        lambda t: t >> group_by(t.col1) >> mutate(any=(C.col1 == C.col2).any()),
    )


def test_op_all(df4):
    assert_result_equal(
        df4,
        lambda t: t >> group_by(t.col1) >> summarise(all=(C.col2 != C.col3).all()),
    )
    assert_result_equal(
        df4,
        lambda t: t >> group_by(t.col1) >> mutate(all=(C.col2 != C.col3).all()),
    )



from __future__ import annotations

from pydiverse.transform import C
from pydiverse.transform.core.verbs import (
    mutate,
    select,
)
from tests.util import assert_result_equal


def test_lambda_cols(df3):
    assert_result_equal(df3, lambda t: t >> select(C.col1, C.col2))
    assert_result_equal(df3, lambda t: t >> mutate(col1=C.col1, col2=C.col1))

    assert_result_equal(df3, lambda t: t >> select(C.col10), exception=ValueError)


def test_columns_pipeable(df3):
    assert_result_equal(df3, lambda t: t.col1 >> mutate(x=t.col1))

    # Test invalid operations
    assert_result_equal(df3, lambda t: t.col1 >> mutate(x=t.col2), exception=ValueError)

    assert_result_equal(df3, lambda t: t.col1 >> mutate(x=C.col2), exception=ValueError)

    assert_result_equal(df3, lambda t: (t.col1 + 1) >> select(), exception=TypeError)



from __future__ import annotations

import pytest

# Setup


supported_options = [
    "postgres",
    "mssql",
    "ibm_db2",
]


def pytest_addoption(parser):
    for opt in supported_options:
        parser.addoption(
            "--" + opt,
            action="store_true",
            default=False,
            help=f"run test that require {opt}",
        )


def pytest_collection_modifyitems(config: pytest.Config, items):
    for opt in supported_options:
        if not config.getoption("--" + opt):
            skip = pytest.mark.skip(reason=f"{opt} not selected")
            for item in items:
                if opt in item.keywords:
                    item.add_marker(skip)



from __future__ import annotations

import sqlite3

import pytest

from pydiverse.transform.core.expressions.lambda_getter import C
from pydiverse.transform.core.verbs import (
    alias,
    join,
    left_join,
    mutate,
    outer_join,
    select,
)
from tests.util import assert_result_equal, full_sort


@pytest.mark.parametrize(
    "how",
    [
        "inner",
        "left",
        pytest.param(
            "outer",
            marks=pytest.mark.skipif(
                sqlite3.sqlite_version < "3.39.0",
                reason="SQLite version doesn't support OUTER JOIN",
            ),
        ),
    ],
)
def test_join(df1, df2, how):
    assert_result_equal(
        (df1, df2),
        lambda t, u: t >> join(u, t.col1 == u.col1, how=how),
    )

    assert_result_equal(
        (df1, df2),
        lambda t, u: t
        >> join(u, (t.col1 == u.col1) & (t.col1 == u.col2), how=how)
        >> mutate(l=t.col2.str.len())
        >> left_join(v := u >> alias("v"), C.l == v.col2)
        >> mutate(k=v.col1 + C.l + u.col2)
        >> outer_join(w := t >> alias("w"), (t.col1 == w.col1) & (C.k == w.col1)),
        check_row_order=False,
    )


@pytest.mark.parametrize(
    "how",
    [
        "inner",
        "left",
        pytest.param(
            "outer",
            marks=pytest.mark.skipif(
                sqlite3.sqlite_version < "3.39.0",
                reason="SQLite version doesn't support OUTER JOIN",
            ),
        ),
    ],
)
def test_join_and_select(df1, df2, how):
    assert_result_equal(
        (df1, df2),
        lambda t, u: t >> select() >> join(u, t.col1 == u.col1, how=how) >> full_sort(),
    )

    assert_result_equal(
        (df1, df2),
        lambda t, u: t
        >> join(u >> select(), (t.col1 == u.col1) & (t.col1 == u.col2), how=how)
        >> full_sort(),
    )


@pytest.mark.parametrize(
    "how",
    [
        "inner",
        "left",
        pytest.param(
            "outer",
            marks=pytest.mark.skipif(
                sqlite3.sqlite_version < "3.39.0",
                reason="SQLite version doesn't support OUTER JOIN",
            ),
        ),
    ],
)
def test_self_join(df3, how):
    # Self join without alias should raise an exception
    assert_result_equal(
        df3,
        lambda t: t >> join(t, t.col1 == t.col1, how=how),
        exception=ValueError,
    )

    def self_join_1(t):
        u = t >> alias("self_join")
        return t >> join(u, t.col1 == u.col1, how=how) >> full_sort()

    assert_result_equal(df3, self_join_1)

    def self_join_2(t):
        u = t >> alias("self_join")
        return (
            t
            >> join(u, (t.col1 == u.col1) & (t.col2 == u.col2), how=how)
            >> full_sort()
        )

    assert_result_equal(df3, self_join_2)

    def self_join_3(t):
        u = t >> alias("self_join")
        return t >> join(u, (t.col2 == u.col3), how=how) >> full_sort()

    assert_result_equal(df3, self_join_3)



from __future__ import annotations

import datetime as dt

import polars as pl
import pytest

from pydiverse.transform import C
from pydiverse.transform.core import dtypes
from pydiverse.transform.core import functions as f
from pydiverse.transform.core.alignment import aligned, eval_aligned
from pydiverse.transform.core.dispatchers import Pipeable, verb
from pydiverse.transform.core.table import Table
from pydiverse.transform.core.verbs import *
from pydiverse.transform.errors import AlignmentError
from pydiverse.transform.polars.polars_table import PolarsEager
from tests.util import assert_equal

df1 = pl.DataFrame(
    {
        "col1": [1, 2, 3, 4],
        "col2": ["a", "b", "c", "d"],
    }
)

df2 = pl.DataFrame(
    {
        "col1": [1, 2, 2, 4, 5, 6],
        "col2": [2, 2, 0, 0, 2, None],
        "col3": [0.0, 0.1, 0.2, 0.3, 0.01, 0.02],
    }
)

df3 = pl.DataFrame(
    {
        "col1": [0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2],
        "col2": [0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1],
        "col3": [0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3],
        "col4": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11],
        "col5": list("abcdefghijkl"),
    }
)

df4 = pl.DataFrame(
    {
        "col1": [None, 0, 0, 0, 0, None, 1, 1, 1, 2, 2, 2, 2],
        "col2": [0, 0, 1, 1, 0, 0, 1, None, 1, 0, 0, 1, 1],
        "col3": [None, None, None, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3],
        "col4": [None, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12],
        "col5": list("abcdefghijkl") + [None],
    }
)

df_left = pl.DataFrame(
    {
        "a": [1, 2, 3, 4],
    }
)

df_right = pl.DataFrame(
    {
        "b": [0, 1, 2, 2],
        "c": [5, 6, 7, 8],
    }
)

df_dt = pl.DataFrame(
    {
        "dt1": [
            dt.datetime(1999, 10, 31, hour=15),
            dt.datetime(2013, 7, 14, second=13, minute=55),
        ],
        "dur1": [dt.timedelta(days=5), dt.timedelta(minutes=4, microseconds=197)],
        "dt2": [dt.datetime(1974, 12, 31), dt.datetime(2020, 2, 14, microsecond=43)],
        "d1": [dt.date(2002, 6, 28), dt.date(2003, 1, 1)],
    }
)


@pytest.fixture(params=["numpy", "arrow"])
def dtype_backend(request):
    return request.param


@pytest.fixture
def tbl1():
    return Table(PolarsEager("df1", df1))


@pytest.fixture
def tbl2():
    return Table(PolarsEager("df2", df2))


@pytest.fixture
def tbl3():
    return Table(PolarsEager("df3", df3))


@pytest.fixture
def tbl4():
    return Table(PolarsEager("df4", df4.clone()))


@pytest.fixture
def tbl_left():
    return Table(PolarsEager("df_left", df_left.clone()))


@pytest.fixture
def tbl_right():
    return Table(PolarsEager("df_right", df_right.clone()))


@pytest.fixture
def tbl_dt():
    return Table(PolarsEager("df_dt", df_dt))


def assert_not_inplace(tbl: Table[PolarsEager], operation: Pipeable):
    """
    Operations should not happen in-place. They should always return a new dataframe.
    """
    initial = tbl._impl.df.clone()
    tbl >> operation
    after = tbl._impl.df

    assert initial.equals(after)


class TestPolarsEager:
    def test_dtype(self, tbl1, tbl2):
        assert isinstance(tbl1.col1._.dtype, dtypes.Int)
        assert isinstance(tbl1.col2._.dtype, dtypes.String)

        assert isinstance(tbl2.col1._.dtype, dtypes.Int)
        assert isinstance(tbl2.col2._.dtype, dtypes.Int)
        assert isinstance(tbl2.col3._.dtype, dtypes.Float)

    def test_build_query(self, tbl1):
        assert (tbl1 >> build_query()) is None

    def test_export(self, tbl1):
        # TODO: test export to other backends
        assert_equal(tbl1, df1)

    def test_select(self, tbl1):
        assert_not_inplace(tbl1, select(tbl1.col1))
        assert_equal(tbl1 >> select(tbl1.col1), df1.select("col1"))
        assert_equal(tbl1 >> select(tbl1.col2), df1.select("col2"))
        assert_equal(tbl1 >> select(), df1.select())

    def test_mutate(self, tbl1):
        assert_not_inplace(tbl1, mutate(x=tbl1.col1))

        assert_equal(
            tbl1 >> mutate(col1times2=tbl1.col1 * 2),
            pl.DataFrame(
                {
                    "col1": [1, 2, 3, 4],
                    "col2": ["a", "b", "c", "d"],
                    "col1times2": [2, 4, 6, 8],
                }
            ),
        )

        assert_equal(
            tbl1 >> select() >> mutate(col1times2=tbl1.col1 * 2),
            pl.DataFrame(
                {
                    "col1times2": [2, 4, 6, 8],
                }
            ),
        )

        # Check proper column referencing
        t = tbl1 >> mutate(col2=tbl1.col1, col1=tbl1.col2) >> select()
        assert_equal(
            t >> mutate(x=t.col1, y=t.col2),
            tbl1 >> select() >> mutate(x=tbl1.col2, y=tbl1.col1),
        )
        assert_equal(
            t >> mutate(x=tbl1.col1, y=tbl1.col2),
            tbl1 >> select() >> mutate(x=tbl1.col1, y=tbl1.col2),
        )

    def test_join(self, tbl_left, tbl_right):
        assert_not_inplace(tbl_left, join(tbl_right, tbl_left.a == tbl_right.b, "left"))

        assert_equal(
            tbl_left
            >> join(tbl_right, tbl_left.a == tbl_right.b, "left")
            >> select(tbl_left.a, tbl_right.b),
            pl.DataFrame({"a": [1, 2, 2, 3, 4], "b_df_right": [1, 2, 2, None, None]}),
            check_row_order=False,
        )

        assert_equal(
            tbl_left
            >> join(tbl_right, tbl_left.a == tbl_right.b, "inner", suffix="")
            >> select(tbl_left.a, tbl_right.b),
            pl.DataFrame({"a": [1, 2, 2], "b": [1, 2, 2]}),
            check_row_order=False,
        )

        assert_equal(
            tbl_left
            >> join(tbl_right, tbl_left.a == tbl_right.b, "outer", suffix="42")
            >> select(tbl_left.a, tbl_right.b),
            pl.DataFrame(
                {
                    "a": [None, 1, 2, 2, 3, 4],
                    "b42": [0, 1, 2, 2, None, None],
                }
            ),
            check_row_order=False,
        )

        # test self-join
        assert_equal(
            tbl_left
            >> inner_join(
                tbl_left2 := tbl_left >> alias(),
                tbl_left.a == tbl_left2.a,
            ),
            df_left.join(df_left, on="a", coalesce=False, suffix="_df_left"),
        )

        assert_equal(
            tbl_right
            >> inner_join(
                tbl_right2 := tbl_right >> alias(), tbl_right.b == tbl_right2.b
            )
            >> inner_join(
                tbl_right3 := tbl_right >> alias(), tbl_right.b == tbl_right3.b
            ),
            df_right.join(df_right, "b", suffix="_df_right", coalesce=False).join(
                df_right, "b", suffix="_df_right1", coalesce=False
            ),
        )

    def test_filter(self, tbl1, tbl2):
        assert_not_inplace(tbl1, filter(tbl1.col1 == 3))

        # Simple filter expressions
        assert_equal(tbl1 >> filter(), df1)
        assert_equal(tbl1 >> filter(tbl1.col1 == tbl1.col1), df1)
        assert_equal(tbl1 >> filter(tbl1.col1 == 3), df1.filter(pl.col("col1") == 3))

        # More complex expressions
        assert_equal(
            tbl1 >> filter(tbl1.col1 // 2 == 1),
            pl.DataFrame({"col1": [2, 3], "col2": ["b", "c"]}),
        )

        assert_equal(
            tbl1 >> filter(1 < tbl1.col1) >> filter(tbl1.col1 < 4),
            df1.filter((1 < pl.col("col1")) & (pl.col("col1") < 4)),
        )

    def test_arrange(self, tbl2, tbl4):
        tbl4.col1.nulls_first()
        assert_not_inplace(tbl2, arrange(tbl2.col2))

        assert_equal(
            tbl2 >> arrange(tbl2.col3) >> select(tbl2.col3),
            df2.select(pl.col("col3")).sort("col3", descending=False),
        )

        assert_equal(
            tbl2 >> arrange(-tbl2.col3) >> select(tbl2.col3),
            df2.select(pl.col("col3")).sort("col3", descending=True),
        )

        assert_equal(
            tbl2 >> arrange(tbl2.col1, tbl2.col2),
            df2.sort(["col1", "col2"], descending=[False, False]),
        )

        assert_equal(
            tbl4
            >> arrange(
                tbl4.col1.nulls_first(),
                -tbl4.col2.nulls_first(),
                -tbl4.col5.nulls_first(),
            ),
            df4.sort(
                ["col1", "col2", "col5"],
                descending=[False, True, True],
                nulls_last=False,
            ),
        )

        assert_equal(
            tbl4
            >> arrange(
                tbl4.col1.nulls_last(),
                -tbl4.col2.nulls_last(),
                -tbl4.col5.nulls_last(),
            ),
            df4.sort(
                ["col1", "col2", "col5"],
                descending=[False, True, True],
                nulls_last=True,
            ),
        )

        assert_equal(
            tbl2 >> arrange(tbl2.col1, tbl2.col2),
            tbl2 >> arrange(tbl2.col2) >> arrange(tbl2.col1),
        )

        assert_equal(tbl2 >> arrange(--tbl2.col3), tbl2 >> arrange(tbl2.col3))  # noqa: B002

    def test_summarise(self, tbl3):
        assert_equal(
            tbl3 >> summarise(mean=tbl3.col1.mean(), max=tbl3.col4.max()),
            pl.DataFrame({"mean": [1.0], "max": [11]}),
            check_row_order=False,
        )

        assert_equal(
            tbl3 >> group_by(tbl3.col1) >> summarise(mean=tbl3.col4.mean()),
            pl.DataFrame({"col1": [0, 1, 2], "mean": [1.5, 5.5, 9.5]}),
            check_row_order=False,
        )

        assert_equal(
            tbl3 >> summarise(mean=tbl3.col4.mean()) >> mutate(mean_2x=C.mean * 2),
            pl.DataFrame({"mean": [5.5], "mean_2x": [11.0]}),
            check_row_order=False,
        )

    def test_group_by(self, tbl3):
        # Grouping doesn't change the result
        assert_equal(tbl3 >> group_by(tbl3.col1), tbl3)
        assert_equal(
            tbl3 >> summarise(mean4=tbl3.col4.mean()) >> group_by(C.mean4),
            tbl3 >> summarise(mean4=tbl3.col4.mean()),
            check_row_order=False,
        )

        # Groupings can be added
        assert_equal(
            tbl3
            >> group_by(tbl3.col1)
            >> group_by(tbl3.col2, add=True)
            >> summarise(mean3=tbl3.col3.mean(), mean4=tbl3.col4.mean()),
            tbl3
            >> group_by(tbl3.col1, tbl3.col2)
            >> summarise(mean3=tbl3.col3.mean(), mean4=tbl3.col4.mean()),
            check_row_order=False,
        )

        # Ungroup doesn't change the result
        assert_equal(
            tbl3
            >> group_by(tbl3.col1)
            >> summarise(mean4=tbl3.col4.mean())
            >> ungroup(),
            tbl3 >> group_by(tbl3.col1) >> summarise(mean4=tbl3.col4.mean()),
            check_row_order=False,
        )

    def test_alias(self, tbl1, tbl2):
        assert_not_inplace(tbl1, alias("tblxxx"))

        x = tbl2 >> alias("x")
        assert x._impl.name == "x"

        # Check that applying alias doesn't change the output
        a = (
            tbl1
            >> mutate(xyz=(tbl1.col1 * tbl1.col1) // 2)
            >> join(tbl2, tbl1.col1 == tbl2.col1, "left", suffix="_right")
            >> mutate(col1=tbl1.col1 - C.xyz)
        )
        b = a >> alias("b")

        assert_equal(a, b)

        # Self Join
        assert_equal(
            tbl2 >> join(x, tbl2.col1 == x.col1, "left", suffix="_right"),
            df2.join(
                df2,
                how="left",
                left_on="col1",
                right_on="col1",
                coalesce=False,
            ),
        )

    def test_window_functions(self, tbl3):
        # Everything else should stay the same
        assert_equal(
            tbl3 >> mutate(x=f.row_number(arrange=[-C.col4])) >> select(*tbl3), df3
        )

        assert_equal(
            (
                tbl3
                >> group_by(C.col2)
                >> select()
                >> mutate(x=f.row_number(arrange=[-C.col4]))
            ),
            pl.DataFrame({"x": [6, 5, 6, 5, 4, 3, 4, 3, 2, 1, 2, 1]}),
        )

        # group_by and partition_by should lead to the same result
        assert_equal(
            (
                tbl3
                >> group_by(C.col2)
                >> select()
                >> mutate(x=f.row_number(arrange=[-C.col4]))
            ),
            (
                tbl3
                >> select()
                >> mutate(x=f.row_number(arrange=[-C.col4], partition_by=[C.col2]))
            ),
        )

    def test_slice_head(self, tbl3):
        @verb
        def slice_head_custom(tbl: Table, n: int, *, offset: int = 0):
            t = (
                tbl
                >> mutate(_n=f.row_number(arrange=[]))
                >> alias()
                >> filter((offset < C._n) & (C._n <= (n + offset)))
            )
            return t >> select(*[c for c in t if c._.name != "_n"])

        assert_equal(
            tbl3 >> slice_head(6),
            tbl3 >> slice_head_custom(6),
        )

        assert_equal(
            tbl3 >> slice_head(3, offset=2),
            tbl3 >> slice_head_custom(3, offset=2),
        )

        assert_equal(
            tbl3 >> slice_head(1, offset=100),
            tbl3 >> slice_head_custom(1, offset=100),
        )

    def test_case_expression(self, tbl3):
        assert_equal(
            (
                tbl3
                >> select()
                >> mutate(
                    col1=C.col1.case(
                        (0, 1),
                        (1, 2),
                        (2, 3),
                        default=-1,
                    )
                )
            ),
            (df3.select("col1") + 1),
        )

        assert_equal(
            (
                tbl3
                >> select()
                >> mutate(
                    x=C.col1.case(
                        (C.col2, 1),
                        (C.col3, 2),
                        default=0,
                    )
                )
            ),
            pl.DataFrame({"x": [1, 1, 0, 0, 0, 2, 1, 1, 0, 0, 2, 0]}),
        )

        assert_equal(
            (
                tbl3
                >> select()
                >> mutate(
                    x=f.case(
                        (C.col1 == C.col2, 1),
                        (C.col1 == C.col3, 2),
                        default=C.col4,
                    )
                )
            ),
            pl.DataFrame({"x": [1, 1, 2, 3, 4, 2, 1, 1, 8, 9, 2, 11]}),
        )

    def test_lambda_column(self, tbl1, tbl2):
        # Select
        assert_equal(tbl1 >> select(C.col1), tbl1 >> select(tbl1.col1))

        # Mutate
        assert_equal(
            tbl1 >> mutate(a=tbl1.col1 * 2) >> select() >> mutate(b=C.a * 2),
            tbl1 >> select() >> mutate(b=tbl1.col1 * 4),
        )

        assert_equal(
            tbl1
            >> mutate(a=tbl1.col1 * 2)
            >> mutate(b=C.a * 2, a=tbl1.col1)
            >> select(C.b),
            tbl1 >> select() >> mutate(b=tbl1.col1 * 4),
        )

        # Join
        assert_equal(
            tbl1
            >> select()
            >> mutate(a=tbl1.col1)
            >> join(tbl2, C.a == tbl2.col1, "left"),
            tbl1
            >> select()
            >> mutate(a=tbl1.col1)
            >> join(tbl2, tbl1.col1 == tbl2.col1, "left"),
        )

        # Join that also uses lambda for the right table
        assert_equal(
            tbl1
            >> select()
            >> mutate(a=tbl1.col1)
            >> join(tbl2, C.a == C.col1_custom_suffix, "left", suffix="_custom_suffix"),
            tbl1
            >> select()
            >> mutate(a=tbl1.col1)
            >> join(tbl2, tbl1.col1 == tbl2.col1, "left", suffix="_custom_suffix"),
        )

        # Filter
        assert_equal(
            tbl1 >> mutate(a=tbl1.col1 * 2) >> filter(C.a % 2 == 0),
            tbl1 >> mutate(a=tbl1.col1 * 2) >> filter((tbl1.col1 * 2) % 2 == 0),
        )

        # Arrange
        assert_equal(
            tbl1 >> mutate(a=tbl1.col1 * 2) >> arrange(C.a),
            tbl1 >> arrange(tbl1.col1) >> mutate(a=tbl1.col1 * 2),
        )

    def test_table_setitem(self, tbl_left, tbl_right):
        tl = tbl_left >> alias("df_left")
        tr = tbl_right >> alias("df_right")

        # Iterate over cols and modify
        for col in tl:
            tl[col] = (col * 2) % 3
        for col in tr:
            tr[col] = (col * 2) % 5

        # Check if it worked...
        assert_equal(
            (tl >> join(tr, C.a == C.b, "left", suffix="")),
            (
                tbl_left
                >> mutate(a=(tbl_left.a * 2) % 3)
                >> join(
                    tbl_right
                    >> mutate(b=(tbl_right.b * 2) % 5, c=(tbl_right.c * 2) % 5),
                    C.a == C.b,
                    "left",
                    suffix="",
                )
            ),
        )

    def test_custom_verb(self, tbl1):
        @verb
        def double_col1(tbl):
            tbl[C.col1] = C.col1 * 2
            return tbl

        # Custom verb should not mutate input object
        assert_not_inplace(tbl1, double_col1())

        assert_equal(tbl1 >> double_col1(), tbl1 >> mutate(col1=C.col1 * 2))

    def test_null(self, tbl4):
        # see if we can compare to null
        assert_equal(
            tbl4 >> mutate(u=tbl4.col1 == tbl4.col3),
            df4.with_columns((pl.col("col1") == pl.col("col3")).alias("u")),
        )

        assert_equal(
            tbl4 >> mutate(u=tbl4.col3.is_null()),
            df4.with_columns(pl.col("col3").is_null().alias("u")),
        )

        # test fill_null
        assert_equal(
            tbl4 >> mutate(u=tbl4.col3.fill_null(tbl4.col2)),
            df4.with_columns(pl.col("col3").fill_null(pl.col("col2")).alias("u")),
        )
        assert_equal(
            tbl4 >> mutate(u=tbl4.col3.fill_null(tbl4.col2)),
            tbl4
            >> mutate(u=f.case((tbl4.col3.is_null(), tbl4.col2), default=tbl4.col3)),
        )

    def test_datetime(self, tbl_dt):
        assert_equal(
            tbl_dt
            >> mutate(
                u=(tbl_dt.dt1 - tbl_dt.dt2),
                v=tbl_dt.d1 - tbl_dt.d1,
                w=(tbl_dt.d1 - tbl_dt.dt2) + tbl_dt.dur1 + dt.timedelta(days=1),
            ),
            df_dt.with_columns(
                (pl.col("dt1") - pl.col("dt2")).alias("u"),
                pl.duration().alias("v"),
                (
                    (pl.col("d1") - pl.col("dt2"))
                    + pl.col("dur1")
                    + pl.lit(dt.timedelta(days=1))
                ).alias("w"),
            ),
        )


class TestPolarsAligned:
    def test_eval_aligned(self, tbl1, tbl3, tbl_left, tbl_right):
        # No exception with correct length
        eval_aligned(tbl_left.a + tbl_left.a)
        eval_aligned(tbl_left.a + tbl_right.b)

        with pytest.raises(AlignmentError):
            eval_aligned(tbl1.col1 + tbl3.col1)

        # Test aggregate functions still work
        eval_aligned(tbl1.col1 + tbl3.col1.mean())

        # Test that `with_` argument gets enforced
        eval_aligned(tbl1.col1 + tbl1.col1, with_=tbl1)
        eval_aligned(tbl_left.a * 2, with_=tbl_left)
        eval_aligned(tbl_left.a * 2, with_=tbl_right)  # Same length
        eval_aligned(
            tbl1.col1.mean(), with_=tbl_left
        )  # Aggregate is aligned with everything

        with pytest.raises(AlignmentError):
            eval_aligned(tbl3.col1 * 2, with_=tbl1)

    def test_aligned_decorator(self, tbl1, tbl3, tbl_left, tbl_right):
        @aligned(with_="a")
        def f(a, b):
            return a + b

        f(tbl3.col1, tbl3.col2)
        f(tbl_left.a, tbl_right.b)

        with pytest.raises(AlignmentError):
            f(tbl1.col1, tbl3.col1)

        # Bad Alignment of return type
        @aligned(with_="a")
        def f(a, b):
            return a.mean() + b

        with pytest.raises(AlignmentError):
            f(tbl1.col1, tbl3.col1)

        # Invalid with_ argument
        with pytest.raises(ValueError):
            aligned(with_="x")(lambda a: 0)

    def test_col_addition(self, tbl_left, tbl_right):
        @aligned(with_="a")
        def f(a, b):
            return a + b

        assert_equal(
            tbl_left >> mutate(x=f(tbl_left.a, tbl_right.b)) >> select(C.x),
            pl.DataFrame({"x": (df_left.get_column("a") + df_right.get_column("b"))}),
        )

        with pytest.raises(AlignmentError):
            f(tbl_left.a, (tbl_right >> filter(C.b == 2)).b)

        with pytest.raises(AlignmentError):
            x = f(tbl_left.a, tbl_right.b)
            tbl_left >> filter(C.a <= 3) >> mutate(x=x)


class TestPrintAndRepr:
    def test_table_str(self, tbl1):
        # Table: df1, backend: PolarsEager
        #    col1 col2
        # 0     1    a
        # 1     2    b
        # 2     3    c
        # 3     4    d

        tbl_str = str(tbl1)

        assert "df1" in tbl_str
        assert "PolarsEager" in tbl_str
        assert str(df1) in tbl_str

    def test_table_repr_html(self, tbl1):
        # jupyter html
        assert "exception" not in tbl1._repr_html_()

    def test_col_str(self, tbl1):
        # Symbolic Expression: <df1.col1(int)>
        # dtype: int
        #
        # 0    1
        # 1    2
        # 2    3
        # 3    4
        # Name: df1_col1_XXXXXXXX, dtype: Int64

        col1_str = str(tbl1.col1)
        series = tbl1._impl.df.get_column(
            tbl1._impl.underlying_col_name[tbl1.col1._.uuid]
        )

        assert str(series) in col1_str
        assert "exception" not in col1_str

    def test_col_html_repr(self, tbl1):
        assert "exception" not in tbl1.col1._repr_html_()

    def test_expr_str(self, tbl1):
        expr_str = str(tbl1.col1 * 2)
        assert "exception" not in expr_str

    def test_expr_html_repr(self, tbl1):
        assert "exception" not in (tbl1.col1 * 2)._repr_html_()

    def test_lambda_str(self, tbl1):
        assert "exception" in str(C.col)
        assert "exception" in str(C.col1 + tbl1.col1)

    def test_eval_expr_str(self, tbl_left, tbl_right):
        valid = tbl_left.a + tbl_right.b
        invalid = tbl_left.a + (tbl_right >> filter(C.b == 2)).b

        assert "exception" not in str(valid)
        assert "exception" in str(invalid)



from __future__ import annotations

from datetime import datetime

from pydiverse.transform import C
from pydiverse.transform.core.verbs import (
    filter,
    mutate,
)
from tests.util import assert_result_equal


def test_eq(df_datetime):
    assert_result_equal(
        df_datetime, lambda t: t >> filter(C.col1 == datetime(1970, 1, 1))
    )
    assert_result_equal(
        df_datetime, lambda t: t >> filter(C.col1 == datetime(2004, 12, 31))
    )
    assert_result_equal(df_datetime, lambda t: t >> filter(C.col1 == C.col2))


def test_nq(df_datetime):
    assert_result_equal(
        df_datetime, lambda t: t >> filter(C.col1 != datetime(1970, 1, 1))
    )
    assert_result_equal(
        df_datetime, lambda t: t >> filter(C.col1 != datetime(2004, 12, 31))
    )
    assert_result_equal(df_datetime, lambda t: t >> filter(C.col1 != C.col2))


def test_lt(df_datetime):
    assert_result_equal(
        df_datetime, lambda t: t >> filter(C.col1 < datetime(1970, 1, 1))
    )
    assert_result_equal(
        df_datetime, lambda t: t >> filter(C.col1 < datetime(2004, 12, 31))
    )
    assert_result_equal(df_datetime, lambda t: t >> filter(C.col1 < C.col2))


def test_gt(df_datetime):
    assert_result_equal(
        df_datetime, lambda t: t >> filter(C.col1 > datetime(1970, 1, 1))
    )
    assert_result_equal(
        df_datetime, lambda t: t >> filter(C.col1 > datetime(2004, 12, 31))
    )
    assert_result_equal(df_datetime, lambda t: t >> filter(C.col1 > C.col2))


def test_le(df_datetime):
    assert_result_equal(df_datetime, lambda t: t >> filter(C.col1 <= C.col2))


def test_ge(df_datetime):
    assert_result_equal(df_datetime, lambda t: t >> filter(C.col1 >= C.col2))


def test_year(df_datetime):
    assert_result_equal(
        df_datetime,
        lambda t: t
        >> mutate(
            x=C.col1.dt.year(),
            y=C.col2.dt.year(),
        ),
    )


def test_month(df_datetime):
    assert_result_equal(
        df_datetime,
        lambda t: t
        >> mutate(
            x=C.col1.dt.month(),
            y=C.col2.dt.month(),
        ),
    )


def test_day(df_datetime):
    assert_result_equal(
        df_datetime,
        lambda t: t
        >> mutate(
            x=C.col1.dt.day(),
            y=C.col2.dt.day(),
        ),
    )


def test_hour(df_datetime):
    assert_result_equal(
        df_datetime,
        lambda t: t
        >> mutate(
            x=C.col1.dt.hour(),
            y=C.col2.dt.hour(),
        ),
    )


def test_minute(df_datetime):
    assert_result_equal(
        df_datetime,
        lambda t: t
        >> mutate(
            x=C.col1.dt.minute(),
            y=C.col2.dt.minute(),
        ),
    )


def test_second(df_datetime):
    assert_result_equal(
        df_datetime,
        lambda t: t
        >> mutate(
            x=C.col1.dt.second(),
            y=C.col2.dt.second(),
        ),
    )


def test_millisecond(df_datetime):
    assert_result_equal(
        df_datetime,
        lambda t: t
        >> mutate(
            x=C.col1.dt.millisecond(),
            y=C.col2.dt.millisecond(),
        ),
    )


def test_day_of_week(df_datetime):
    assert_result_equal(
        df_datetime,
        lambda t: t
        >> mutate(
            x=C.col1.dt.day_of_week(),
            y=C.col2.dt.day_of_week(),
        ),
    )


def test_day_of_year(df_datetime):
    assert_result_equal(
        df_datetime,
        lambda t: t
        >> mutate(
            x=C.col1.dt.day_of_year(),
            y=C.col2.dt.day_of_year(),
        ),
    )



from __future__ import annotations

from pydiverse.transform import C
from pydiverse.transform.core import functions as f
from pydiverse.transform.core.verbs import (
    arrange,
    filter,
    group_by,
    mutate,
    select,
    summarise,
    ungroup,
)
from pydiverse.transform.errors import FunctionTypeError
from tests.util import assert_result_equal, full_sort


def test_simple_ungrouped(df3):
    assert_result_equal(
        df3,
        lambda t: t >> mutate(min=t.col4.min(), max=t.col4.max(), mean=t.col4.mean()),
    )


def test_simple_grouped(df3):
    assert_result_equal(
        df3,
        lambda t: t
        >> group_by(t.col1)
        >> mutate(min=t.col4.min(), max=t.col4.max(), mean=t.col4.mean()),
    )

    assert_result_equal(
        df3,
        lambda t: t
        >> group_by(t.col1, t.col2)
        >> mutate(min=t.col4.min(), max=t.col4.max(), mean=t.col4.mean()),
    )


def test_partition_by_argument(df3):
    assert_result_equal(
        df3,
        lambda t: t
        >> mutate(
            u=t.col1.min(partition_by=[t.col3]),
            v=t.col4.sum(partition_by=[t.col2]),
            w=f.rank(arrange=[-t.col5, t.col4], partition_by=[t.col2]),
            x=f.row_number(
                arrange=[t.col4.nulls_last()], partition_by=[t.col1, t.col2]
            ),
        ),
    )

    assert_result_equal(
        df3,
        lambda t: t
        >> group_by(t.col1)
        >> mutate(
            u=t.col3.sum(),
            v=t.col2.sum(partition_by=[t.col2]),
        ),
    )


def test_chained(df3):
    assert_result_equal(
        df3,
        lambda t: t
        >> group_by(t.col1)
        >> mutate(min=t.col4.min())
        >> mutate(max=t.col4.max(), mean=t.col4.mean()),
    )

    assert_result_equal(
        df3,
        lambda t: t
        >> group_by(t.col1)
        >> mutate(min=t.col4.min(), max=t.col4.max())
        >> mutate(span=C.max - C.min),
    )


def test_nested(df3):
    assert_result_equal(
        df3,
        lambda t: t
        >> group_by(t.col1)
        >> mutate(range=t.col4.max() - 10)
        >> ungroup()
        >> mutate(range_mean=C.range.mean()),
    )

    assert_result_equal(
        df3,
        lambda t: t
        >> mutate(x=C.col4.max())
        >> mutate(y=C.x.min() * 1)
        >> mutate(z=C.y.mean())
        >> mutate(w=C.x / C.y),
    )

    assert_result_equal(
        df3,
        lambda t: t >> mutate(x=(C.col4.max().min() + C.col2.mean()).max()),
        exception=FunctionTypeError,
        may_throw=True,
    )


def test_filter(df3):
    assert_result_equal(
        df3,
        lambda t: t
        >> group_by(t.col1, t.col2)
        >> mutate(mean3=t.col3.mean())
        >> filter(C.mean3 <= 2.0),
    )


# def test_filter_argument(df3, df4):
#     assert_result_equal(
#         df4, lambda t: t >> mutate(u=t.col2.mean(filter=~t.col2.is_null()))
#     )

#     assert_result_equal(
#         df4, lambda t: t >> mutate(u=t.col2.mean(filter=~(t.col4 % 3 == 0)))
#     )

#     assert_result_equal(
#         df3, lambda t: t >> mutate(u=t.col4.sum(partition_by=[t.col2]))
#     )

#     assert_result_equal(
#         df4,
#         lambda t: t
#         >> mutate(
#             u=t.col1.min(filter=(~t.col1.is_null()), partition_by=[t.col3]),
#             v=t.col4.max(filter=~t.col4.is_null(), partition_by=[t.col1]),
#         ),
#     )

#     assert_result_equal(
#         df4, lambda t: t >> mutate(u=t.col3.min(filter=t.col3.is_null()))
#     )


def test_arrange(df3):
    assert_result_equal(
        df3,
        lambda t: t
        >> group_by(t.col1, t.col2)
        >> mutate(mean3=t.col3.mean())
        >> arrange(C.mean3),
    )

    assert_result_equal(
        df3,
        lambda t: t
        >> arrange(-t.col4)
        >> group_by(t.col1, t.col2)
        >> mutate(mean3=t.col3.mean())
        >> arrange(C.mean3),
    )


def test_summarise(df3):
    assert_result_equal(
        df3,
        lambda t: t
        >> group_by(t.col1, t.col2)
        >> mutate(range=t.col4.max() - t.col4.min())
        >> summarise(mean_range=C.range.mean()),
    )

    assert_result_equal(
        df3,
        lambda t: t
        >> group_by(t.col1, t.col2)
        >> summarise(range=t.col4.max() - t.col4.min())
        >> mutate(mean_range=C.range.mean()),
    )


def test_intermediate_select(df3):
    # Check that subqueries happen transparently
    assert_result_equal(
        df3,
        lambda t: t
        >> group_by(t.col1, t.col2)
        >> mutate(x=t.col4.mean())
        >> select()
        >> mutate(y=C.x.min())
        >> select()
        >> mutate(z=(C.x - C.y).mean()),
    )


def test_arrange_argument(df3):
    # Grouped
    assert_result_equal(
        df3,
        lambda t: t
        >> group_by(t.col1)
        >> mutate(x=C.col4.shift(1, arrange=[-C.col3]))
        >> full_sort()
        >> select(C.x),
    )

    assert_result_equal(
        df3,
        lambda t: t
        >> group_by(t.col2)
        >> mutate(x=f.row_number(arrange=[-C.col4]))
        >> full_sort()
        >> select(C.x),
    )

    # Ungrouped
    assert_result_equal(
        df3,
        lambda t: t
        >> mutate(x=C.col4.shift(1, arrange=[-C.col3]))
        >> full_sort()
        >> select(C.x),
    )

    assert_result_equal(
        df3,
        lambda t: t
        >> mutate(x=f.row_number(arrange=[-C.col4]))
        >> full_sort()
        >> select(C.x),
    )


def test_complex(df3):
    # Window function before summarise
    assert_result_equal(
        df3,
        lambda t: t
        >> group_by(t.col1, t.col2)
        >> mutate(mean3=t.col3.mean(), rn=f.row_number(arrange=[C.col1, C.col2]))
        >> filter(C.mean3 > C.rn)
        >> summarise(meta_mean=C.mean3.mean())
        >> filter(t.col1 >= C.meta_mean)
        >> filter(t.col1 != 1)
        >> arrange(C.meta_mean),
    )

    # Window function after summarise
    assert_result_equal(
        df3,
        lambda t: t
        >> group_by(t.col1, t.col2)
        >> summarise(mean3=t.col3.mean())
        >> mutate(minM3=C.mean3.min(), maxM3=C.mean3.max())
        >> mutate(span=C.maxM3 - C.minM3)
        >> filter(C.span < 3)
        >> arrange(C.span),
    )


def test_nested_bool(df4):
    assert_result_equal(
        df4,
        lambda t: t
        >> group_by(t.col1)
        >> mutate(x=t.col1 <= t.col2, y=(t.col3 * 4) >= C.col4)
        >> mutate(
            xshift=C.x.shift(1, arrange=[t.col4]),
            yshift=C.y.shift(-1, arrange=[t.col4]),
        )
        >> mutate(xAndY=C.x & C.y, xAndYshifted=C.xshift & C.yshift),
    )


# Test specific operations


def test_op_shift(df4):
    assert_result_equal(
        df4,
        lambda t: t
        >> group_by(t.col1)
        >> mutate(
            shift1=t.col2.shift(1, arrange=[t.col4]),
            shift2=t.col4.shift(-2, 0, arrange=[t.col4]),
            shift3=t.col4.shift(0, arrange=[t.col4]),
            u=C.col1.shift(1, 0, arrange=[t.col4]),
        ),
    )

    assert_result_equal(
        df4,
        lambda t: t
        >> mutate(
            u=t.col1.shift(1, 0, arrange=[t.col2, t.col4]),
            v=t.col1.shift(2, 1, arrange=[-t.col4.nulls_first()]),
        ),
    )


def test_op_row_number(df4):
    assert_result_equal(
        df4,
        lambda t: t
        >> group_by(t.col1)
        >> mutate(
            row_number1=f.row_number(arrange=[-C.col4.nulls_last()]),
            row_number2=f.row_number(arrange=[C.col2, C.col3, t.col4]),
        ),
    )

    assert_result_equal(
        df4,
        lambda t: t
        >> mutate(
            u=f.row_number(arrange=[-C.col4.nulls_last()]),
            v=f.row_number(arrange=[-t.col3, t.col4]),
        ),
    )


def test_op_rank(df4):
    assert_result_equal(
        df4,
        lambda t: t
        >> group_by(t.col1)
        >> mutate(
            rank1=f.rank(arrange=[t.col1]),
            rank2=f.rank(arrange=[t.col2]),
            rank3=f.rank(arrange=[t.col2.nulls_last()]),
            rank4=f.rank(arrange=[t.col5.nulls_first()]),
            rank5=f.rank(arrange=[-t.col5.nulls_first()]),
            rank_expr=f.rank(arrange=[t.col3 - t.col2]),
        ),
    )


def test_op_dense_rank(df3):
    assert_result_equal(
        df3,
        lambda t: t
        >> group_by(t.col1)
        >> mutate(
            rank1=f.dense_rank(arrange=[t.col5.nulls_first()]),
            rank2=f.dense_rank(arrange=[t.col2]),
            rank3=f.dense_rank(arrange=[t.col2.nulls_last()]),
        )
        >> ungroup(),
        # TODO: activate these once SQL partition_by= is implemented
        # >> mutate(
        #    rank4=f.dense_rank(arrange=[t.col4.nulls_first()], partition_by=[t.col2]),
        #    rank5=f.dense_rank(arrange=[-t.col5.nulls_first()], partition_by=[t.col2]),
        # ),
    )



from __future__ import annotations

from pydiverse.transform.ops.core import Nullary, Window

__all__ = [
    "Shift",
    "RowNumber",
    "Rank",
    "DenseRank",
]


class Shift(Window):
    name = "shift"
    signatures = [
        "T, const int -> T",
        "T, const int, const T -> T",
    ]


class RowNumber(Window, Nullary):
    name = "row_number"
    signatures = [
        "-> int",
    ]


class Rank(Window, Nullary):
    name = "rank"
    signatures = [
        "-> int",
    ]


class DenseRank(Window, Nullary):
    name = "dense_rank"
    signatures = [
        "-> int",
    ]



from __future__ import annotations

from pydiverse.transform.ops.core import ElementWise, OperatorExtension, Unary
from pydiverse.transform.ops.numeric import Add, RAdd, RSub, Sub

__all__ = [
    "DtYear",
    "DtMonth",
    "DtDay",
    "DtHour",
    "DtMinute",
    "DtSecond",
    "DtMillisecond",
    "DtDayOfWeek",
    "DtDayOfYear",
    "DtDays",
    "DtHours",
    "DtMinutes",
    "DtSeconds",
    "DtMilliseconds",
    "DtSub",
    "DtRSub",
    "DtDurAdd",
    "DtDurRAdd",
]


class DtExtract(ElementWise, Unary):
    signatures = ["datetime -> int"]


class DateExtract(ElementWise, Unary):
    signatures = ["date -> int"]


class DtYear(DtExtract, DateExtract):
    name = "dt.year"


class DtMonth(DtExtract, DateExtract):
    name = "dt.month"


class DtDay(DtExtract, DateExtract):
    name = "dt.day"


class DtHour(DtExtract):
    name = "dt.hour"


class DtMinute(DtExtract):
    name = "dt.minute"


class DtSecond(DtExtract):
    name = "dt.second"


class DtMillisecond(DtExtract):
    name = "dt.millisecond"


class DtDayOfWeek(DtExtract, DateExtract):
    name = "dt.day_of_week"


class DtDayOfYear(DtExtract, DateExtract):
    name = "dt.day_of_year"


class DurationToUnit(ElementWise, Unary):
    signatures = ["duration -> int"]


class DtDays(DurationToUnit):
    name = "dt.days"


class DtHours(DurationToUnit):
    name = "dt.hours"


class DtMinutes(DurationToUnit):
    name = "dt.minutes"


class DtSeconds(DurationToUnit):
    name = "dt.seconds"


class DtMilliseconds(DurationToUnit):
    name = "dt.milliseconds"


class DtSub(OperatorExtension):
    operator = Sub
    signatures = [
        "datetime, datetime -> duration",
        "datetime, date -> duration",
        "date, datetime -> duration",
        "date, date -> duration",
    ]


class DtRSub(OperatorExtension):
    operator = RSub
    signatures = [
        "datetime, datetime -> duration",
        "datetime, date -> duration",
        "date, datetime -> duration",
        "date, date -> duration",
    ]


class DtDurAdd(OperatorExtension):
    operator = Add
    signatures = ["duration, duration -> duration"]


class DtDurRAdd(OperatorExtension):
    operator = RAdd
    signatures = ["duration, duration -> duration"]



from __future__ import annotations

from datetime import datetime

import polars as pl
import pytest

from tests.fixtures.backend import BACKEND_MARKS, flatten
from tests.util.backend import BACKEND_TABLES

dataframes = {
    "df1": pl.DataFrame(
        {
            "col1": [1, 2, 3, 4],
            "col2": ["a", "baa", "c", "d"],
        }
    ),
    "df2": pl.DataFrame(
        {
            "col1": [1, 2, -2, 4, 5, 6],
            "col2": [2, 2, 0, 0, 2, None],
            "col3": [0.0, -0.1, 0.2, 0.3, 0.01, 0.02],
        }
    ),
    "df3": pl.DataFrame(
        {  # tests rely on col4 having a unique ordering
            "col1": [0, 0, -1, -1000, 1, 1, 0, 1, 4, 2, 2, 2],
            "col2": [0, 0, 10, 1, 0, 5, 1, 1, 0, 0, 1, 1],
            "col3": [0, 1, 2, 3, 0, 1, 2, 3, 1, 1, 2, 3],
            "col4": [-1729, 2, 1, 3, 4, 5, 13, 7, 8, 9, 10, 11],
            "col5": list("abcdafghijkk"),
        }
    ),
    "df4": pl.DataFrame(
        {  # tests rely on col4 having a unique ordering
            "col1": [None, 0, 0, 0, 0, None, 1, 1, 1, 2, 3, 2, 2],
            "col2": [0, 20, 1, 1, 0, 0, 1, None, 1, 0, 0, 0, 1],
            "col3": [None, None, None, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3],
            "col4": [None, -42, 2, 33, 4, 5, -10, 7, 8, 90, 10, 11, 12],
            "col5": list("abcdefghijkl") + [None],
        }
    ),
    "df_strings": pl.DataFrame(
        {
            "col1": [
                "",
                " ",
                "xyzzzz123",
                " x ",
                "foo",
                "FooBarfooofoo",
                "abracadabra",
                None,
                "--+011x",
                None,
                "_ %",
            ],
            "col2": [
                "",
                "test_%",
                "xyz",
                " x",
                "FOO33",
                "FooBar",
                "AbracadabA",
                "",
                "barAb",
                None,
                "% _.AbAbAb",
            ],
        }
    ),
    "df_datetime": pl.DataFrame(
        {
            "col1": [
                datetime(2000, 1, 1),
                datetime(2000, 1, 1, 12, 15, 59),
                datetime(1970, 1, 1, 12, 30, 30, 987_123),
                datetime(1700, 4, 27, 1, 2, 3),
                None,
                datetime(2250, 12, 24, 23, 0, 0),
                datetime(2023, 7, 31, 10, 16, 13),
                datetime(2004, 12, 31),
                datetime(1970, 1, 1),
            ],
            "col2": [
                datetime(2023, 7, 31, 10, 16, 13),
                datetime(2023, 7, 31, 10, 16, 14),
                datetime(1900, 1, 2, 3, 4, 5),
                datetime(1988, 5, 3, 12, 4, 54),
                None,
                datetime(2023, 7, 31, 10, 16, 13),
                None,
                datetime(2004, 12, 31, 23, 59, 59, 456_789),
                datetime(1970, 1, 1),
            ],
        }
    ),
}

# compare one dataframe and one SQL backend to all others
# (some tests are ignored if either backend does not support a feature)
reference_backends = ["polars", "duckdb"]


def pytest_generate_tests(metafunc: pytest.Metafunc):
    pl.Config.set_tbl_cols(12)
    pl.Config.set_tbl_rows(40)
    # Parametrize tests based on `backends` and `skip_backends` mark.

    backends = dict.fromkeys(BACKEND_TABLES)
    # if mark := metafunc.definition.get_closest_marker("backends"):
    #     backends = dict.fromkeys(mark.args)
    if mark := metafunc.definition.get_closest_marker("skip_backends"):
        for backend in mark.args:
            if backend in backends:
                del backends[backend]

    backends = {k: i for i, k in enumerate(backends)}

    backend_combinations = [
        (reference_backend, backend)
        for reference_backend in reference_backends
        for backend in backends
    ]

    params = []
    table_names = [name for name in metafunc.fixturenames if name in dataframes]
    for reference_backend, backend in backend_combinations:
        # Skip some redundant backend combinations
        if reference_backend == backend:
            continue
        if reference_backend not in backends:
            continue
        if backend in reference_backends:
            if backends[backend] < backends[reference_backend]:
                continue

        marks = list(
            flatten(
                [
                    BACKEND_MARKS.get(reference_backend, ()),
                    BACKEND_MARKS.get(backend, ()),
                ]
            )
        )

        arguments = [(reference_backend, backend)] * len(table_names)

        params.append(
            pytest.param(
                *arguments,
                marks=marks,
                id=f"{reference_backend}-{backend}",
            )
        )

    param_names = ",".join([f"{name}" for name in table_names])
    return metafunc.parametrize(param_names, params, indirect=True)


def generate_df_fixtures():
    def gen_fixture(table_name):
        @pytest.fixture(scope="function")
        def table_fixture(request):
            df = dataframes[table_name]
            name = f"equiv_{table_name}"

            table_x = BACKEND_TABLES[request.param[0]](df, name)
            table_y = BACKEND_TABLES[request.param[1]](df, name)
            return table_x, table_y

        return table_fixture

    for name in dataframes:
        globals()[f"{name}"] = gen_fixture(name)


generate_df_fixtures()



from __future__ import annotations

import functools

import polars as pl

from pydiverse.transform.core import Table
from pydiverse.transform.polars.polars_table import PolarsEager
from pydiverse.transform.sql.sql_table import SQLTableImpl


def _cached_impl(fn):
    cache = {}

    @functools.wraps(fn)
    def wrapped(df: pl.DataFrame, name: str):
        if name in cache:
            return cache[name]

        impl = fn(df, name)
        cache[name] = impl

        return impl

    return wrapped


@_cached_impl
def polars_impl(df: pl.DataFrame, name: str):
    return PolarsEager(name, df)


_sql_engine_cache = {}


def _sql_table(df: pl.DataFrame, name: str, url: str, dtypes_map: dict = None):
    import sqlalchemy as sa

    global _sql_engine_cache

    dtypes_map = dtypes_map or {}

    if url in _sql_engine_cache:
        engine = _sql_engine_cache[url]
    else:
        engine = sa.create_engine(url)
        _sql_engine_cache[url] = engine

    sql_dtypes = {}
    for col, dtype in zip(df.columns, df.dtypes):
        if dtype in dtypes_map:
            sql_dtypes[col] = dtypes_map[dtype]

    df.write_database(
        name, engine, if_table_exists="replace", engine_options={"dtype": sql_dtypes}
    )
    return SQLTableImpl(engine, name)


@_cached_impl
def sqlite_impl(df: pl.DataFrame, name: str):
    return _sql_table(df, name, "sqlite:///:memory:")


@_cached_impl
def duckdb_impl(df: pl.DataFrame, name: str):
    return _sql_table(df, name, "duckdb:///:memory:")


@_cached_impl
def postgres_impl(df: pl.DataFrame, name: str):
    url = "postgresql://sa:Pydiverse23@127.0.0.1:6543"
    return _sql_table(df, name, url)


@_cached_impl
def mssql_impl(df: pl.DataFrame, name: str):
    from sqlalchemy.dialects.mssql import DATETIME2

    url = (
        "mssql+pyodbc://sa:PydiQuant27@127.0.0.1:1433"
        "/master?driver=ODBC+Driver+18+for+SQL+Server&encrypt=no"
    )
    return _sql_table(
        df,
        name,
        url,
        dtypes_map={
            pl.Datetime(): DATETIME2(),
        },
    )


def impl_to_table_callable(fn):
    @functools.wraps(fn)
    def wrapped(df: pl.DataFrame, name: str):
        impl = fn(df, name)
        return Table(impl)

    return wrapped


BACKEND_TABLES = {
    "polars": impl_to_table_callable(polars_impl),
    "sqlite": impl_to_table_callable(sqlite_impl),
    "duckdb": impl_to_table_callable(duckdb_impl),
    "postgres": impl_to_table_callable(postgres_impl),
    "mssql": impl_to_table_callable(mssql_impl),
}



from __future__ import annotations

from pydiverse.transform import C
from pydiverse.transform.core.verbs import (
    filter,
    mutate,
)
from tests.util import assert_result_equal


def test_eq(df_strings):
    assert_result_equal(df_strings, lambda t: t >> filter(C.col1 == " "))
    assert_result_equal(df_strings, lambda t: t >> filter(C.col1 == "foo"))
    assert_result_equal(df_strings, lambda t: t >> filter(C.col1 == C.col2))


def test_nq(df_strings):
    assert_result_equal(df_strings, lambda t: t >> filter(C.col1 != " "))
    assert_result_equal(df_strings, lambda t: t >> filter(C.col1 != "foo"))
    assert_result_equal(df_strings, lambda t: t >> filter(C.col1 != C.col2))


def test_lt(df_strings):
    assert_result_equal(df_strings, lambda t: t >> filter(C.col1 < " x"))
    assert_result_equal(df_strings, lambda t: t >> filter(C.col1 < "E"))
    assert_result_equal(df_strings, lambda t: t >> filter(C.col1 < C.col2))


def test_gt(df_strings):
    assert_result_equal(df_strings, lambda t: t >> filter(C.col1 > " x"))
    assert_result_equal(df_strings, lambda t: t >> filter(C.col1 > "E"))
    assert_result_equal(df_strings, lambda t: t >> filter(C.col1 > C.col2))


def test_le(df_strings):
    assert_result_equal(df_strings, lambda t: t >> filter(C.col1 <= C.col2))


def test_ge(df_strings):
    assert_result_equal(df_strings, lambda t: t >> filter(C.col1 >= C.col2))


def test_strip(df_strings):
    assert_result_equal(
        df_strings,
        lambda t: t
        >> mutate(
            x=C.col1.str.strip(),
            y=C.col2.str.strip(),
        ),
    )


def test_string_length(df_strings):
    assert_result_equal(
        df_strings,
        lambda t: t
        >> mutate(
            x=C.col1.str.len(),
            y=C.col2.str.len(),
        ),
    )


def test_to_upper(df_strings):
    assert_result_equal(
        df_strings,
        lambda t: t
        >> mutate(
            x=C.col1.str.to_upper(),
            y=C.col2.str.to_upper(),
        ),
    )


def test_to_lower(df_strings):
    assert_result_equal(
        df_strings,
        lambda t: t
        >> mutate(
            x=C.col1.str.to_lower(),
            y=C.col2.str.to_lower(),
        ),
    )


def test_replace_all(df_strings):
    assert_result_equal(
        df_strings,
        lambda t: t
        >> mutate(
            x=C.col1.str.replace_all(" ", "").str.replace_all("foo", "fOO"),
            y=C.col2.str.replace_all("Ab", "ab"),
        ),
    )


def test_starts_with(df_strings):
    assert_result_equal(
        df_strings,
        lambda t: t
        >> mutate(
            x=C.col1.str.starts_with("foo") | C.col1.str.starts_with(" "),
            y=C.col2.str.starts_with("test") | C.col2.str.starts_with("Abra"),
            underscore=C.col1.str.starts_with("_"),
            percent=C.col2.str.starts_with("%"),
        ),
    )


def test_ends_with(df_strings):
    assert_result_equal(
        df_strings,
        lambda t: t
        >> mutate(
            x=C.col1.str.ends_with("Bar") | C.col1.str.ends_with(" "),
            y=C.col2.str.ends_with("_%") | C.col2.str.ends_with("Bar"),
            percent=C.col1.str.ends_with("%"),
            underscore=C.col2.str.ends_with("_"),
        ),
    )


def test_contains(df_strings):
    assert_result_equal(
        df_strings,
        lambda t: t
        >> mutate(
            x=C.col1.str.contains(" ") | C.col1.str.contains("Foo"),
            y=C.col2.str.contains("st_") | C.col2.str.contains("bar"),
            percent=C.col1.str.contains("%"),
            underscore=C.col2.str.contains("_"),
        ),
    )


def test_slice(df_strings):
    assert_result_equal(
        df_strings,
        lambda t: t
        >> mutate(
            u=t.col1.str.slice(1, 3),
            v=t.col2.str.slice(t.col1.str.len() % (t.col2.str.len() + 1), 42),
            w=t.col1.str.slice(2, t.col1.str.len()),
        ),
    )



from __future__ import annotations

from pydiverse.transform import C
from pydiverse.transform import functions as f
from pydiverse.transform.core.verbs import mutate
from tests.fixtures.backend import skip_backends
from tests.util import assert_result_equal


def test_count(df4):
    assert_result_equal(
        df4, lambda t: t >> mutate(**{col._.name + "_count": f.count(col) for col in t})
    )


def test_row_number(df4):
    assert_result_equal(
        df4,
        lambda t: t >> mutate(row_number=f.row_number(arrange=[-C.col1, C.col5])),
    )


# MSSQL Added the LEAST function in version 2022.
# Our docker container doesn't yet support it.
@skip_backends("mssql")
def test_min(df4):
    assert_result_equal(
        df4,
        lambda t: t
        >> mutate(
            int1=f.min(C.col1 + 2, C.col2, 9),
            int2=f.min(C.col1 * C.col2, 0),
            int3=f.min(C.col1 * C.col2, C.col2 * C.col3, 2 - C.col3),
            int4=f.min(C.col1),
            float1=f.min(C.col1, 1.5),
            float2=f.min(1, C.col1 + 1.5, C.col2, 2.2),
            str1=f.min(C.col5, "c"),
            str2=f.min(C.col5, "C"),
        ),
    )


# MSSQL Added the GREATEST function in version 2022.
# Our docker container doesn't yet support it.
@skip_backends("mssql")
def test_max(df4):
    assert_result_equal(
        df4,
        lambda t: t
        >> mutate(
            int1=f.max(C.col1 + 2, C.col2, 9),
            int2=f.max(C.col1 * C.col2, 0),
            int3=f.max(C.col1 * C.col2, C.col2 * C.col3, 2 - C.col3),
            int4=f.max(C.col1),
            float1=f.max(C.col1, 1.5),
            float2=f.max(1, C.col1 + 1.5, C.col2, 2.2),
            str1=f.max(C.col5, "c"),
            str2=f.max(C.col5, "C"),
        ),
    )



from __future__ import annotations

import sqlite3

import polars as pl
import pytest
import sqlalchemy as sa

from pydiverse.transform import C
from pydiverse.transform.core import functions as f
from pydiverse.transform.core.alignment import aligned, eval_aligned
from pydiverse.transform.core.table import Table
from pydiverse.transform.core.verbs import *
from pydiverse.transform.errors import AlignmentError
from pydiverse.transform.sql.sql_table import SQLTableImpl
from tests.util import assert_equal

df1 = pl.DataFrame(
    {
        "col1": [1, 2, 3, 4],
        "col2": ["a", "b", "c", "d"],
    }
)

df2 = pl.DataFrame(
    {
        "col1": [1, 2, 2, 4, 5, 6],
        "col2": [2, 2, 0, 0, 2, None],
        "col3": [0.0, 0.1, 0.2, 0.3, 0.01, 0.02],
    }
)

df3 = pl.DataFrame(
    {
        "col1": [0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2],
        "col2": [0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1],
        "col3": [0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3],
        "col4": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11],
        "col5": list("abcdefghijkl"),
    }
)

df4 = pl.DataFrame(
    {
        "col1": [None, 0, 0, 0, 0, None, 1, 1, 1, 2, 2, 2, 2],
        "col2": [0, 0, 1, 1, 0, 0, 1, None, 1, 0, 0, 1, 1],
        "col3": [None, None, None, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3],
        "col4": [None, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12],
        "col5": list("abcdefghijkl") + [None],
    }
)

df_left = pl.DataFrame(
    {
        "a": [1, 2, 3, 4],
    }
)

df_right = pl.DataFrame(
    {
        "b": [0, 1, 2, 2],
        "c": [5, 6, 7, 8],
    }
)


@pytest.fixture
def engine():
    engine = sa.create_engine("sqlite:///:memory:")
    df1.write_database("df1", engine, if_table_exists="replace")
    df2.write_database("df2", engine, if_table_exists="replace")
    df3.write_database("df3", engine, if_table_exists="replace")
    df4.write_database("df4", engine, if_table_exists="replace")
    df_left.write_database("df_left", engine, if_table_exists="replace")
    df_right.write_database("df_right", engine, if_table_exists="replace")
    return engine


@pytest.fixture
def tbl1(engine):
    return Table(SQLTableImpl(engine, "df1"))


@pytest.fixture
def tbl2(engine):
    return Table(SQLTableImpl(engine, "df2"))


@pytest.fixture
def tbl3(engine):
    return Table(SQLTableImpl(engine, "df3"))


@pytest.fixture
def tbl4(engine):
    return Table(SQLTableImpl(engine, "df4"))


@pytest.fixture
def tbl_left(engine):
    return Table(SQLTableImpl(engine, "df_left"))


@pytest.fixture
def tbl_right(engine):
    return Table(SQLTableImpl(engine, "df_right"))


class TestSQLTable:
    def test_build_query(self, tbl1):
        query_str = tbl1 >> build_query()
        expected_out = "SELECT df1.col1 AS col1, df1.col2 AS col2 FROM df1"
        assert query_str.lower().split() == expected_out.lower().split()

    def test_show_query(self, tbl1, capfd):
        tbl1 >> show_query()
        out = capfd.readouterr().out
        expected_out = "SELECT df1.col1 AS col1, df1.col2 AS col2 FROM df1"

        assert out.lower().split() == expected_out.lower().split()

        # Verify that it is chainable
        tbl1 >> show_query() >> collect()

    def test_export(self, tbl1):
        assert_equal(tbl1 >> export(), df1)

    def test_select(self, tbl1, tbl2):
        assert_equal(tbl1 >> select(tbl1.col1), df1[["col1"]])
        assert_equal(tbl1 >> select(tbl1.col2), df1[["col2"]])

    def test_mutate(self, tbl1):
        assert_equal(
            tbl1 >> mutate(col1times2=tbl1.col1 * 2),
            pl.DataFrame(
                {
                    "col1": [1, 2, 3, 4],
                    "col2": ["a", "b", "c", "d"],
                    "col1times2": [2, 4, 6, 8],
                }
            ),
        )

        assert_equal(
            tbl1 >> select() >> mutate(col1times2=tbl1.col1 * 2),
            pl.DataFrame(
                {
                    "col1times2": [2, 4, 6, 8],
                }
            ),
        )

        # Check proper column referencing
        t = tbl1 >> mutate(col2=tbl1.col1, col1=tbl1.col2) >> select()
        assert_equal(
            t >> mutate(x=t.col1, y=t.col2),
            tbl1 >> select() >> mutate(x=tbl1.col2, y=tbl1.col1),
        )
        assert_equal(
            t >> mutate(x=tbl1.col1, y=tbl1.col2),
            tbl1 >> select() >> mutate(x=tbl1.col1, y=tbl1.col2),
        )

    def test_join(self, tbl_left, tbl_right):
        assert_equal(
            tbl_left
            >> join(tbl_right, tbl_left.a == tbl_right.b, "left", suffix="")
            >> select(tbl_left.a, tbl_right.b),
            pl.DataFrame({"a": [1, 2, 2, 3, 4], "b": [1, 2, 2, None, None]}),
        )

        assert_equal(
            tbl_left
            >> join(tbl_right, tbl_left.a == tbl_right.b, "inner", suffix="")
            >> select(tbl_left.a, tbl_right.b),
            pl.DataFrame({"a": [1, 2, 2], "b": [1, 2, 2]}),
        )

        if sqlite3.sqlite_version_info >= (3, 39, 0):
            assert_equal(
                (
                    tbl_left
                    >> join(
                        tbl_right, tbl_left.a == tbl_right.b, "outer", suffix="_1729"
                    )
                    >> select(tbl_left.a, tbl_right.b)
                ),
                pl.DataFrame(
                    {
                        "a": [1.0, 2.0, 2.0, 3.0, 4.0, None],
                        "b_1729": [1.0, 2.0, 2.0, None, None, 0.0],
                    }
                ),
            )

    def test_filter(self, tbl1, tbl2):
        # Simple filter expressions
        assert_equal(tbl1 >> filter(), df1)
        assert_equal(tbl1 >> filter(tbl1.col1 == tbl1.col1), df1)
        assert_equal(tbl1 >> filter(tbl1.col1 == 3), df1.filter(pl.col("col1") == 3))

        # More complex expressions
        assert_equal(
            tbl1 >> filter(tbl1.col1 // 2 == 1),
            pl.DataFrame({"col1": [2, 3], "col2": ["b", "c"]}),
        )

        assert_equal(
            tbl1 >> filter(1 < tbl1.col1) >> filter(tbl1.col1 < 4),
            df1.filter((1 < pl.col("col1")) & (pl.col("col1") < 4)),
        )

    def test_arrange(self, tbl2):
        assert_equal(
            tbl2 >> arrange(tbl2.col3) >> select(tbl2.col3),
            df2.select("col3").sort("col3", descending=False),
        )

        assert_equal(
            tbl2 >> arrange(-tbl2.col3) >> select(tbl2.col3),
            df2.select("col3").sort("col3", descending=True),
        )

        assert_equal(
            tbl2 >> arrange(tbl2.col1, tbl2.col2),
            df2.sort(["col1", "col2"], descending=[False, False]),
        )

        assert_equal(
            tbl2 >> arrange(tbl2.col1, tbl2.col2),
            tbl2 >> arrange(tbl2.col2) >> arrange(tbl2.col1),
        )

        assert_equal(tbl2 >> arrange(--tbl2.col3), tbl2 >> arrange(tbl2.col3))  # noqa: B002

    def test_summarise(self, tbl3):
        assert_equal(
            tbl3 >> summarise(mean=tbl3.col1.mean(), max=tbl3.col4.max()),
            pl.DataFrame({"mean": [1], "max": [11]}),
        )

        assert_equal(
            tbl3 >> group_by(tbl3.col1) >> summarise(mean=tbl3.col4.mean()),
            pl.DataFrame({"col1": [0, 1, 2], "mean": [1.5, 5.5, 9.5]}),
        )

        assert_equal(
            tbl3 >> summarise(mean=tbl3.col4.mean()) >> mutate(mean_2x=C.mean * 2),
            pl.DataFrame({"mean": [5.5], "mean_2x": [11]}),
        )

    def test_group_by(self, tbl3):
        # Grouping doesn't change the result
        assert_equal(tbl3 >> group_by(tbl3.col1), tbl3)
        assert_equal(
            tbl3 >> summarise(mean4=tbl3.col4.mean()) >> group_by(C.mean4),
            tbl3 >> summarise(mean4=tbl3.col4.mean()),
        )

        # Groupings can be added
        assert_equal(
            tbl3
            >> group_by(tbl3.col1)
            >> group_by(tbl3.col2, add=True)
            >> summarise(mean3=tbl3.col3.mean(), mean4=tbl3.col4.mean()),
            tbl3
            >> group_by(tbl3.col1, tbl3.col2)
            >> summarise(mean3=tbl3.col3.mean(), mean4=tbl3.col4.mean()),
        )

        # Ungroup doesn't change the result
        assert_equal(
            tbl3
            >> group_by(tbl3.col1)
            >> summarise(mean4=tbl3.col4.mean())
            >> ungroup(),
            tbl3 >> group_by(tbl3.col1) >> summarise(mean4=tbl3.col4.mean()),
        )

    def test_alias(self, tbl1, tbl2):
        x = tbl2 >> alias("x")
        assert x._impl.name == "x"

        # Check that applying alias doesn't change the output
        a = (
            tbl1
            >> mutate(xyz=(tbl1.col1 * tbl1.col1) // 2)
            >> join(tbl2, tbl1.col1 == tbl2.col1, "left")
            >> mutate(col1=tbl1.col1 - C.xyz)
        )
        b = a >> alias("b")

        assert_equal(a, b)

        # Self Join
        self_join = (
            tbl2
            >> join(x, tbl2.col1 == x.col1, "left", suffix="42")
            >> alias("self_join")
        )
        self_join >>= arrange(*self_join)

        self_join_expected = df2.join(
            df2,
            how="left",
            left_on="col1",
            right_on="col1",
            coalesce=False,
            suffix="42",
        )
        self_join_expected = self_join_expected.sort(
            by=[col._.name for col in self_join]
        )

        assert_equal(self_join, self_join_expected)

    def test_lambda_column(self, tbl1, tbl2):
        # Select
        assert_equal(tbl1 >> select(C.col1), tbl1 >> select(tbl1.col1))

        # Mutate
        assert_equal(
            tbl1 >> mutate(a=tbl1.col1 * 2) >> select() >> mutate(b=C.a * 2),
            tbl1 >> select() >> mutate(b=tbl1.col1 * 4),
        )

        assert_equal(
            tbl1
            >> mutate(a=tbl1.col1 * 2)
            >> mutate(b=C.a * 2, a=tbl1.col1)
            >> select(C.b),
            tbl1 >> select() >> mutate(b=tbl1.col1 * 4),
        )

        # Join
        assert_equal(
            tbl1
            >> select()
            >> mutate(a=tbl1.col1 * 2)
            >> join(tbl2, C.a == tbl2.col1, "left"),
            tbl1
            >> select()
            >> mutate(a=tbl1.col1 * 2)
            >> join(tbl2, tbl1.col1 * 2 == tbl2.col1, "left"),
        )

        # Join that also uses lambda for the right table
        assert_equal(
            tbl1
            >> select()
            >> mutate(a=tbl1.col1)
            >> join(tbl2, C.a == C.col1_df2, "left"),
            tbl1
            >> select()
            >> mutate(a=tbl1.col1)
            >> join(tbl2, tbl1.col1 == tbl2.col1, "left"),
        )

        # Filter
        assert_equal(
            tbl1 >> mutate(a=tbl1.col1 * 2) >> filter(C.a % 2 == 0),
            tbl1 >> mutate(a=tbl1.col1 * 2) >> filter((tbl1.col1 * 2) % 2 == 0),
        )

        # Arrange
        assert_equal(
            tbl1 >> mutate(a=tbl1.col1 * 2) >> arrange(C.a),
            tbl1 >> arrange(tbl1.col1) >> mutate(a=tbl1.col1 * 2),
        )

    def test_table_setitem(self, tbl_left, tbl_right):
        tl = tbl_left >> alias("df_left")
        tr = tbl_right >> alias("df_right")

        # Iterate over cols and modify
        for col in tl:
            tl[col] = (col * 2) % 3
        for col in tr:
            tr[col] = (col * 2) % 5

        # Check if it worked...
        assert_equal(
            (tl >> join(tr, C.a == tr.b, "left")),
            (
                tbl_left
                >> mutate(a=(tbl_left.a * 2) % 3)
                >> join(
                    tbl_right
                    >> mutate(b=(tbl_right.b * 2) % 5, c=(tbl_right.c * 2) % 5),
                    C.a == C.b_df_right,
                    "left",
                )
            ),
        )

    def test_select_without_tbl_ref(self, tbl2):
        assert_equal(
            tbl2 >> summarise(count=f.count()),
            tbl2 >> summarise(count=f.count(tbl2.col1)),
        )

        assert_equal(
            tbl2 >> summarise(count=f.count()), pl.DataFrame({"count": [len(df2)]})
        )

    def test_null_comparison(self, tbl4):
        assert_equal(
            tbl4 >> mutate(u=tbl4.col1 == tbl4.col3),
            df4.with_columns((pl.col("col1") == pl.col("col3")).alias("u")),
        )

        assert_equal(
            tbl4 >> mutate(u=tbl4.col3.is_null()),
            df4.with_columns(pl.col("col3").is_null().alias("u")),
        )


class TestSQLAligned:
    def test_eval_aligned(self, tbl1, tbl3, tbl_left, tbl_right):
        # Columns must be from same table
        eval_aligned(tbl_left.a + tbl_left.a)
        eval_aligned(tbl3.col1 + tbl3.col2)

        # Derived columns are also OK
        tbl1_mutate = tbl1 >> mutate(x=tbl1.col1 * 2)
        eval_aligned(tbl1.col1 + tbl1_mutate.x)

        with pytest.raises(AlignmentError):
            eval_aligned(tbl1.col1 + tbl3.col1)
        with pytest.raises(AlignmentError):
            eval_aligned(tbl_left.a + tbl_right.b)
        with pytest.raises(AlignmentError):
            eval_aligned(tbl1.col1 + tbl3.col1.mean())
        with pytest.raises(AlignmentError):
            tbl1_joined = tbl1 >> join(tbl3, tbl1.col1 == tbl3.col1, how="left")
            eval_aligned(tbl1.col1 + tbl1_joined.col1)

        # Test that `with_` argument gets enforced
        eval_aligned(tbl1.col1 + tbl1.col1, with_=tbl1)
        eval_aligned(tbl_left.a * 2, with_=tbl_left)
        eval_aligned(tbl1.col1, with_=tbl1_mutate)

        with pytest.raises(AlignmentError):
            eval_aligned(tbl1.col1.mean(), with_=tbl_left)

        with pytest.raises(AlignmentError):
            eval_aligned(tbl3.col1 * 2, with_=tbl1)

        with pytest.raises(AlignmentError):
            eval_aligned(tbl_left.a, with_=tbl_right)

    def test_aligned_decorator(self, tbl1, tbl3, tbl_left, tbl_right):
        @aligned(with_="a")
        def f(a, b):
            return a + b

        f(tbl3.col1, tbl3.col2)
        f(tbl_right.b, tbl_right.c)

        with pytest.raises(AlignmentError):
            f(tbl1.col1, tbl3.col1)

        with pytest.raises(AlignmentError):
            f(tbl_left.a, tbl_right.b)

        # Check with_ parameter gets enforced
        @aligned(with_="a")
        def f(a, b):
            return b

        f(tbl1.col1, tbl1.col2)
        with pytest.raises(AlignmentError):
            f(tbl1.col1, tbl3.col1)

        # Invalid with_ argument
        with pytest.raises(ValueError):
            aligned(with_="x")(lambda a: 0)

    def test_col_addition(self, tbl3):
        @aligned(with_="a")
        def f(a, b):
            return a + b

        assert_equal(
            tbl3 >> mutate(x=f(tbl3.col1, tbl3.col2)) >> select(C.x),
            pl.DataFrame({"x": [0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3]}),
        )

        # Test if it also works with derived tables
        tbl3_mutate = tbl3 >> mutate(x=tbl3.col1 * 2)
        tbl3 >> mutate(x=f(tbl3_mutate.col1, tbl3_mutate.x))

        with pytest.raises(AlignmentError):
            tbl3 >> arrange(C.col1) >> mutate(x=f(tbl3.col1, tbl3.col2))

        with pytest.raises(AlignmentError):
            tbl3 >> filter(C.col1 == 1) >> mutate(x=f(tbl3.col1, tbl3.col2))



from __future__ import annotations

import functools
import itertools
import operator
import uuid
from typing import Any, Callable, Literal

import polars as pl

from pydiverse.transform import ops
from pydiverse.transform.core import dtypes
from pydiverse.transform.core.expressions.expressions import (
    BaseExpression,
    CaseExpression,
    Column,
    FunctionCall,
    LiteralColumn,
)
from pydiverse.transform.core.expressions.symbolic_expressions import SymbolicExpression
from pydiverse.transform.core.expressions.translator import (
    Translator,
    TypedValue,
)
from pydiverse.transform.core.registry import TypedOperatorImpl
from pydiverse.transform.core.table_impl import AbstractTableImpl
from pydiverse.transform.core.util import OrderingDescriptor
from pydiverse.transform.core.util.util import translate_ordering
from pydiverse.transform.errors import (
    AlignmentError,
    ExpressionError,
    FunctionTypeError,
)
from pydiverse.transform.ops.core import OPType


class PolarsEager(AbstractTableImpl):
    def __init__(self, name: str, df: pl.DataFrame):
        self.df = df
        self.join_translator = JoinTranslator()

        cols = {
            col.name: Column(col.name, self, _pdt_dtype(col.dtype))
            for col in df.iter_columns()
        }
        self.underlying_col_name: dict[uuid.UUID, str] = {
            col.uuid: f"{name}_{col.name}_{col.uuid.int}" for col in cols.values()
        }
        self.df = self.df.rename(
            {col.name: self.underlying_col_name[col.uuid] for col in cols.values()}
        )
        super().__init__(name, cols)

    def mutate(self, **kwargs):
        uuid_to_kwarg: dict[uuid.UUID, (str, BaseExpression)] = {
            self.named_cols.fwd[k]: (k, v) for (k, v) in kwargs.items()
        }
        self.underlying_col_name.update(
            {
                uuid: f"{self.name}_{col_name}_mut_{uuid.int}"
                for uuid, (col_name, _) in uuid_to_kwarg.items()
            }
        )

        polars_exprs = [
            self.cols[uuid].compiled().alias(self.underlying_col_name[uuid])
            for uuid in uuid_to_kwarg.keys()
        ]
        self.df = self.df.with_columns(*polars_exprs)

    def join(
        self,
        right: PolarsEager,
        on: SymbolicExpression,
        how: Literal["inner", "left", "outer"],
        *,
        validate: Literal["1:1", "1:m", "m:1", "m:m"] = "m:m",
    ):
        # get the columns on which the data frames are joined
        left_on: list[str] = []
        right_on: list[str] = []
        for col1, col2 in self.join_translator.translate(on):
            if col2.uuid in self.cols and col1.uuid in right.cols:
                col1, col2 = col2, col1
            assert col1.uuid in self.cols and col2.uuid in right.cols
            left_on.append(self.underlying_col_name[col1.uuid])
            right_on.append(right.underlying_col_name[col2.uuid])

        self.underlying_col_name.update(right.underlying_col_name)

        self.df = self.df.join(
            right.df,
            how=how,
            left_on=left_on,
            right_on=right_on,
            validate=validate,
            coalesce=False,
        )

    def filter(self, *args: SymbolicExpression):
        if not args:
            return
        pl_expr, dtype = self.compiler.translate(functools.reduce(operator.and_, args))
        assert isinstance(dtype, dtypes.Bool)
        self.df = self.df.filter(pl_expr())

    def alias(self, new_name: str | None = None):
        new_name = new_name or self.name
        return self.__class__(new_name, self.export())

    def arrange(self, ordering: list[OrderingDescriptor]):
        self.df = self.df.sort(
            by=[self.compiler.translate(o.order).value() for o in ordering],
            nulls_last=[not o.nulls_first for o in ordering],
            descending=[not o.asc for o in ordering],
        )

    def summarise(self, **kwargs: SymbolicExpression):
        uuid_to_kwarg: dict[uuid.UUID, (str, BaseExpression)] = {
            self.named_cols.fwd[k]: (k, v) for (k, v) in kwargs.items()
        }
        self.underlying_col_name.update(
            {
                uuid: f"{self.name}_{col_name}_summarise_{uuid.int}"
                for uuid, (col_name, _) in uuid_to_kwarg.items()
            }
        )

        agg_exprs: list[pl.Expr] = [
            self.cols[uuid].compiled().alias(self.underlying_col_name[uuid])
            for uuid in uuid_to_kwarg.keys()
        ]
        group_exprs: list[pl.Expr] = [
            pl.col(self.underlying_col_name[col.uuid]) for col in self.grouped_by
        ]

        if self.grouped_by:
            # retain the cols the table was grouped by and add the aggregation cols
            self.df = self.df.group_by(*group_exprs).agg(*agg_exprs)
        else:
            self.df = self.df.select(*agg_exprs)

    def export(self) -> pl.DataFrame:
        return self.df.select(
            **{
                name: self.underlying_col_name[uuid]
                for (name, uuid) in self.selected_cols()
            }
        )

    def slice_head(self, n: int, offset: int):
        self.df = self.df.slice(offset, n)

    def is_aligned_with(self, col: Column | LiteralColumn) -> bool:
        if isinstance(col, Column):
            return (
                isinstance(col.table, type(self))
                and col.table.df.height == self.df.height
            )
        if isinstance(col, LiteralColumn):
            return issubclass(col.backend, type(self)) and (
                not isinstance(col.typed_value.value, pl.Series)
                or len(col.typed_value.value) == self.df.height
            )  # not a series => scalar

    class ExpressionCompiler(
        AbstractTableImpl.ExpressionCompiler[
            "PolarsEager", TypedValue[Callable[[], pl.Expr]]
        ]
    ):
        def _translate_col(
            self, col: Column, **kwargs
        ) -> TypedValue[Callable[[], pl.Expr]]:
            def value():
                return pl.col(self.backend.underlying_col_name[col.uuid])

            return TypedValue(value, col.dtype)

        def _translate_literal_col(
            self, col: LiteralColumn, **kwargs
        ) -> TypedValue[Callable[[], pl.Expr]]:
            if not self.backend.is_aligned_with(col):
                raise AlignmentError(
                    f"literal column {col} not aligned with table {self.backend.name}."
                )

            def value(**kw):
                return col.typed_value.value

            return TypedValue(value, col.typed_value.dtype, col.typed_value.ftype)

        def _translate_function(
            self,
            implementation: TypedOperatorImpl,
            op_args: list[TypedValue[Callable[[], pl.Expr]]],
            context_kwargs: dict[str, Any],
            *,
            verb: str | None = None,
            **kwargs,
        ) -> TypedValue[Callable[[], pl.Expr]]:
            pl_result_type = _pl_dtype(implementation.rtype)

            internal_kwargs = {}

            op = implementation.operator
            ftype = (
                OPType.WINDOW
                if op.ftype == OPType.AGGREGATE and verb != "summarise"
                else op.ftype
            )

            grouping = context_kwargs.get("partition_by")
            # the `partition_by=` grouping overrides the `group_by` grouping
            if grouping is not None:  # translate possible lambda cols
                grouping = [self.backend.resolve_lambda_cols(col) for col in grouping]
            else:  # use the current grouping of the table
                grouping = self.backend.grouped_by

            ordering = context_kwargs.get("arrange")
            if ordering:
                ordering = translate_ordering(self.backend, ordering)
                by = [self._translate(o.order).value() for o in ordering]
                descending = [not o.asc for o in ordering]
                nulls_last = [not o.nulls_first for o in ordering]

            filter_cond = context_kwargs.get("filter")
            if filter_cond:
                filter_cond = self.translate(
                    self.backend.resolve_lambda_cols(filter_cond)
                )

            args: list[Callable[[], pl.Expr]] = [arg.value for arg in op_args]
            dtypes: list[dtypes.DType] = [arg.dtype for arg in op_args]
            if ftype == OPType.WINDOW and ordering and not grouping:
                # order the args. if the table is grouped by group_by or
                # partition_by=, the groups will be sorted via over(order_by=)
                # anyways so it need not be done here.
                def ordered_arg(arg):
                    return arg().sort_by(
                        by=by, descending=descending, nulls_last=nulls_last
                    )

                args = [
                    arg if dtype.const else functools.partial(ordered_arg, arg)
                    for arg, dtype in zip(args, dtypes)
                ]

            if ftype in (OPType.WINDOW, OPType.AGGREGATE) and filter_cond:
                # filtering needs to be done before applying the operator. We filter
                # all non-constant arguments, although there should always be only
                # one of these.
                def filtered_value(value):
                    return value().filter(filter_cond.value())

                assert len(list(filter(lambda arg: not arg.dtype.const, op_args))) == 1
                args = [
                    arg if dtype.const else functools.partial(filtered_value, arg)
                    for arg, dtype in zip(args, dtypes)
                ]

            if op.name in ("rank", "dense_rank"):
                assert len(args) == 0
                args = [
                    functools.partial(
                        lambda ordering: pl.struct(
                            *self.backend._merge_desc_nulls_last(ordering)
                        ),
                        ordering,
                    )
                ]
                ordering = None

            def value(**kw):
                return implementation(
                    *[arg(**kw) for arg in args],
                    _tbl=self.backend,
                    _result_type=pl_result_type,
                    **internal_kwargs,
                )

            if ftype == OPType.AGGREGATE:
                if context_kwargs.get("filter"):
                    # TODO: allow AGGRRGATE + `filter` context_kwarg
                    raise NotImplementedError

                if context_kwargs.get("partition_by"):
                    # technically, it probably wouldn't be too hard to support this in
                    # polars.
                    assert verb == "summarise"
                    raise ValueError(
                        f"cannot use keyword argument `partition_by` for the "
                        f"aggregation function `{op.name}` inside `summarise`."
                    )

            # TODO: in the grouping / filter expressions, we should probably call
            # validate_table_args. look what it does and use it.
            # TODO: what happens if I put None or similar in a filter / partition_by?
            if ftype == OPType.WINDOW:
                if verb == "summarise":
                    raise FunctionTypeError(
                        "window function are not allowed inside summarise"
                    )

                # if `verb` != "muatate", we should give a warning that this only works
                # for polars

                if grouping:
                    # when doing sort_by -> over in polars, for whatever reason the
                    # `nulls_last` argument is ignored. thus when both a grouping and an
                    # arrangment are specified, we manually add the descending and
                    # nulls_last markers to the ordering.
                    order_by = None
                    if ordering:
                        order_by = self.backend._merge_desc_nulls_last(ordering)

                    def partitioned_value(value):
                        group_exprs: list[pl.Expr] = [
                            pl.col(self.backend.underlying_col_name[col.uuid])
                            for col in grouping
                        ]
                        return value().over(*group_exprs, order_by=order_by)

                    value = functools.partial(partitioned_value, value)

                elif ordering:
                    if op.ftype == OPType.AGGREGATE:
                        # TODO: don't fail, but give a warning that `arrange` is useless
                        # here
                        ...

                    # the function was executed on the ordered arguments. here we
                    # restore the original order of the table.
                    def sorted_value(value):
                        inv_permutation = pl.int_range(
                            0, pl.len(), dtype=pl.Int64
                        ).sort_by(
                            by=by,
                            descending=descending,
                            nulls_last=nulls_last,
                        )
                        return value().sort_by(inv_permutation)

                    # need to bind `value` inside `filtered_value` so that it refers to
                    # the original `value`.
                    value = functools.partial(sorted_value, value)

            return TypedValue(
                value,
                implementation.rtype,
                PolarsEager._get_op_ftype(
                    op_args,
                    op,
                    OPType.WINDOW
                    if op.ftype == OPType.AGGREGATE and verb != "summarise"
                    else None,
                ),
            )

        def _translate_case(
            self,
            expr: CaseExpression,
            switching_on: TypedValue[Callable[[], pl.Expr]] | None,
            cases: list[
                tuple[
                    TypedValue[Callable[[], pl.Expr]], TypedValue[Callable[[], pl.Expr]]
                ]
            ],
            default: TypedValue[Callable[[], pl.Expr]],
            **kwargs,
        ) -> TypedValue[Callable[[], pl.Expr]]:
            def value():
                if switching_on is not None:
                    switching_on_v = switching_on.value()
                    conds = [
                        match_expr.value() == switching_on_v for match_expr, _ in cases
                    ]
                else:
                    conds = [case[0].value() for case in cases]

                pl_expr = pl.when(conds[0]).then(cases[0][1].value())
                for cond, (_, value) in zip(conds[1:], cases[1:]):
                    pl_expr = pl_expr.when(cond).then(value.value())
                return pl_expr.otherwise(default.value())

            result_dtype, result_ftype = self._translate_case_common(
                expr, switching_on, cases, default, **kwargs
            )

            return TypedValue(value, result_dtype, result_ftype)

        def _translate_literal_value(self, expr):
            def value():
                return pl.lit(expr)

            return value

    class AlignedExpressionEvaluator(
        AbstractTableImpl.AlignedExpressionEvaluator[TypedValue[pl.Series]]
    ):
        def _translate_col(self, col: Column, **kwargs) -> TypedValue[pl.Series]:
            return TypedValue(
                col.table.df.get_column(col.table.underlying_col_name[col.uuid]),
                col.table.cols[col.uuid].dtype,
            )

        def _translate_literal_col(
            self, expr: LiteralColumn, **kwargs
        ) -> TypedValue[pl.Series]:
            return expr.typed_value

        def _translate_function(
            self,
            implementation: TypedOperatorImpl,
            op_args: list[TypedValue[pl.Series]],
            context_kwargs: dict[str, Any],
            **kwargs,
        ) -> TypedValue[pl.Series]:
            args = [arg.value for arg in op_args]
            op = implementation.operator

            arg_lens = {arg.len() for arg in args if isinstance(arg, pl.Series)}
            if len(arg_lens) >= 2:
                raise AlignmentError(
                    f"arguments for function {implementation.operator.name} are not "
                    f"aligned. they have lengths {list(arg_lens)} but all lengths must "
                    f"be equal."
                )

            value = implementation(*args)

            return TypedValue(
                value,
                implementation.rtype,
                PolarsEager._get_op_ftype(
                    op_args, op, OPType.WINDOW if op.ftype == OPType.AGGREGATE else None
                ),
            )

    # merges descending and null_last markers into the ordering expression
    def _merge_desc_nulls_last(
        self, ordering: list[OrderingDescriptor]
    ) -> list[pl.Expr]:
        with_signs = []
        for o in ordering:
            numeric = (
                self.compiler.translate(o.order).value().rank("dense").cast(pl.Int64)
            )
            with_signs.append(numeric if o.asc else -numeric)
        return [
            x.fill_null(
                -(pl.len().cast(pl.Int64) + 1)
                if o.nulls_first
                else pl.len().cast(pl.Int64) + 1
            )
            for x, o in zip(with_signs, ordering)
        ]


class JoinTranslator(Translator[tuple]):
    """
    This translator takes a conjunction (AND) of equality checks and returns
    a tuple of tuple where the inner tuple contains the left and right column
    of the equality checks.
    """

    def _translate(self, expr, **kwargs):
        if isinstance(expr, Column):
            return expr
        if isinstance(expr, FunctionCall):
            if expr.name == "__eq__":
                c1 = expr.args[0]
                c2 = expr.args[1]
                assert isinstance(c1, Column) and isinstance(c2, Column)
                return ((c1, c2),)
            if expr.name == "__and__":
                return tuple(itertools.chain(*expr.args))
        raise ExpressionError(
            f"invalid ON clause element: {expr}. only a conjunction of equalities"
            " is supported"
        )


def _pdt_dtype(t: pl.DataType) -> dtypes.DType:
    if t.is_float():
        return dtypes.Float()
    elif t.is_integer():
        return dtypes.Int()
    elif isinstance(t, pl.Boolean):
        return dtypes.Bool()
    elif isinstance(t, pl.String):
        return dtypes.String()
    elif isinstance(t, pl.Datetime):
        return dtypes.DateTime()
    elif isinstance(t, pl.Date):
        return dtypes.Date()
    elif isinstance(t, pl.Duration):
        return dtypes.Duration()

    raise TypeError(f"polars type {t} is not supported")


def _pl_dtype(t: dtypes.DType) -> pl.DataType:
    if isinstance(t, dtypes.Float):
        return pl.Float64()
    elif isinstance(t, dtypes.Int):
        return pl.Int64()
    elif isinstance(t, dtypes.Bool):
        return pl.Boolean()
    elif isinstance(t, dtypes.String):
        return pl.String()
    elif isinstance(t, dtypes.DateTime):
        return pl.Datetime()
    elif isinstance(t, dtypes.Date):
        return pl.Date()
    elif isinstance(t, dtypes.Duration):
        return pl.Duration()

    raise TypeError(f"pydiverse.transform type {t} not supported for polars")


with PolarsEager.op(ops.Mean()) as op:

    @op.auto
    def _mean(x):
        return x.mean()


with PolarsEager.op(ops.Min()) as op:

    @op.auto
    def _min(x):
        return x.min()


with PolarsEager.op(ops.Max()) as op:

    @op.auto
    def _max(x):
        return x.max()


with PolarsEager.op(ops.Sum()) as op:

    @op.auto
    def _sum(x):
        return x.sum()


with PolarsEager.op(ops.All()) as op:

    @op.auto
    def _all(x):
        return x.all()


with PolarsEager.op(ops.Any()) as op:

    @op.auto
    def _any(x):
        return x.any()


with PolarsEager.op(ops.IsNull()) as op:

    @op.auto
    def _is_null(x):
        return x.is_null()


with PolarsEager.op(ops.IsNotNull()) as op:

    @op.auto
    def _is_not_null(x):
        return x.is_not_null()


with PolarsEager.op(ops.FillNull()) as op:

    @op.auto
    def _fill_null(x, y):
        return x.fill_null(y)


with PolarsEager.op(ops.DtYear()) as op:

    @op.auto
    def _dt_year(x):
        return x.dt.year()


with PolarsEager.op(ops.DtMonth()) as op:

    @op.auto
    def _dt_month(x):
        return x.dt.month()


with PolarsEager.op(ops.DtDay()) as op:

    @op.auto
    def _dt_day(x):
        return x.dt.day()


with PolarsEager.op(ops.DtHour()) as op:

    @op.auto
    def _dt_hour(x):
        return x.dt.hour()


with PolarsEager.op(ops.DtMinute()) as op:

    @op.auto
    def _dt_minute(x):
        return x.dt.minute()


with PolarsEager.op(ops.DtSecond()) as op:

    @op.auto
    def _dt_second(x):
        return x.dt.second()


with PolarsEager.op(ops.DtMillisecond()) as op:

    @op.auto
    def _dt_millisecond(x):
        return x.dt.millisecond()


with PolarsEager.op(ops.DtDayOfWeek()) as op:

    @op.auto
    def _dt_day_of_week(x):
        return x.dt.weekday()


with PolarsEager.op(ops.DtDayOfYear()) as op:

    @op.auto
    def _dt_day_of_year(x):
        return x.dt.ordinal_day()


with PolarsEager.op(ops.DtDays()) as op:

    @op.auto
    def _days(x):
        return x.dt.total_days()


with PolarsEager.op(ops.DtHours()) as op:

    @op.auto
    def _hours(x):
        return x.dt.total_hours()


with PolarsEager.op(ops.DtMinutes()) as op:

    @op.auto
    def _minutes(x):
        return x.dt.total_minutes()


with PolarsEager.op(ops.DtSeconds()) as op:

    @op.auto
    def _seconds(x):
        return x.dt.total_seconds()


with PolarsEager.op(ops.DtMilliseconds()) as op:

    @op.auto
    def _milliseconds(x):
        return x.dt.total_milliseconds()


with PolarsEager.op(ops.Sub()) as op:

    @op.extension(ops.DtSub)
    def _dt_sub(lhs, rhs):
        return lhs - rhs


with PolarsEager.op(ops.RSub()) as op:

    @op.extension(ops.DtRSub)
    def _dt_rsub(rhs, lhs):
        return lhs - rhs


with PolarsEager.op(ops.Add()) as op:

    @op.extension(ops.DtDurAdd)
    def _dt_dur_add(lhs, rhs):
        return lhs + rhs


with PolarsEager.op(ops.RAdd()) as op:

    @op.extension(ops.DtDurRAdd)
    def _dt_dur_radd(rhs, lhs):
        return lhs + rhs


with PolarsEager.op(ops.RowNumber()) as op:

    @op.auto
    def _row_number():
        return pl.int_range(start=1, end=pl.len() + 1, dtype=pl.Int64)


with PolarsEager.op(ops.Rank()) as op:

    @op.auto
    def _rank(x):
        return x.rank("min").cast(pl.Int64)


with PolarsEager.op(ops.DenseRank()) as op:

    @op.auto
    def _dense_rank(x):
        return x.rank("dense").cast(pl.Int64)


with PolarsEager.op(ops.Shift()) as op:

    @op.auto
    def _shift(x, n, fill_value=None):
        return x.shift(n, fill_value=fill_value)


with PolarsEager.op(ops.IsIn()) as op:

    @op.auto
    def _isin(x, *values):
        return x.is_in([pl.select(v).item() for v in values])


with PolarsEager.op(ops.StrContains()) as op:

    @op.auto
    def _contains(x, y):
        return x.str.contains(y)


with PolarsEager.op(ops.StrStartsWith()) as op:

    @op.auto
    def _starts_with(x, y):
        return x.str.starts_with(y)


with PolarsEager.op(ops.StrEndsWith()) as op:

    @op.auto
    def _ends_with(x, y):
        return x.str.ends_with(y)


with PolarsEager.op(ops.StrToLower()) as op:

    @op.auto
    def _lower(x):
        return x.str.to_lowercase()


with PolarsEager.op(ops.StrToUpper()) as op:

    @op.auto
    def _upper(x):
        return x.str.to_uppercase()


with PolarsEager.op(ops.StrReplaceAll()) as op:

    @op.auto
    def _replace_all(x, to_replace, replacement):
        return x.str.replace_all(to_replace, replacement)


with PolarsEager.op(ops.StrLen()) as op:

    @op.auto
    def _string_length(x):
        return x.str.len_chars().cast(pl.Int64)


with PolarsEager.op(ops.StrStrip()) as op:

    @op.auto
    def _str_strip(x):
        return x.str.strip_chars()


with PolarsEager.op(ops.StrSlice()) as op:

    @op.auto
    def _str_slice(x, offset, length):
        return x.str.slice(offset, length)


with PolarsEager.op(ops.Count()) as op:

    @op.auto
    def _count(x=None):
        return pl.len() if x is None else x.count()


with PolarsEager.op(ops.Greatest()) as op:

    @op.auto
    def _greatest(*x):
        return pl.max_horizontal(*x)


with PolarsEager.op(ops.Least()) as op:

    @op.auto
    def _least(*x):
        return pl.min_horizontal(*x)



from __future__ import annotations

from pydiverse.transform.core.expressions import LambdaColumn
from pydiverse.transform.core.expressions.symbolic_expressions import SymbolicExpression

__all__ = ["C"]


class MC(type):
    def __getattr__(cls, name: str) -> SymbolicExpression:
        return SymbolicExpression(LambdaColumn(name))

    def __getitem__(cls, name: str) -> SymbolicExpression:
        return SymbolicExpression(LambdaColumn(name))


class C(metaclass=MC):
    pass



from __future__ import annotations

from pydiverse.transform import C
from pydiverse.transform.core.verbs import (
    filter,
    mutate,
)
from tests.util import assert_result_equal


def test_noop(df2):
    assert_result_equal(df2, lambda t: t >> filter())
    assert_result_equal(df2, lambda t: t >> filter(t.col1 == t.col1))


def test_simple_filter(df2):
    assert_result_equal(df2, lambda t: t >> filter(t.col1 == 2))
    assert_result_equal(df2, lambda t: t >> filter(t.col1 != 2))


def test_chained_filters(df2):
    assert_result_equal(df2, lambda t: t >> filter(1 < t.col1) >> filter(t.col1 < 5))

    assert_result_equal(df2, lambda t: t >> filter(1 < t.col1) >> filter(t.col3 < 0.25))


def test_filter_empty_result(df3):
    assert_result_equal(
        df3,
        lambda t: t >> filter(t.col1 == 0) >> filter(t.col2 == 2) >> filter(t.col4 < 2),
    )


def test_filter_after_mutate(df4):
    assert_result_equal(
        df4,
        lambda t: t >> mutate(x=t.col1 <= t.col2) >> filter(C.x),
    )


def test_filter_isin(df4):
    assert_result_equal(
        df4,
        lambda t: t
        >> filter(
            C.col1.isin(0, 2),
        ),
    )

    assert_result_equal(
        df4,
        lambda t: t
        >> mutate(x=t.col1.isin(0, 2))
        >> filter(
            t.col2.isin(0, 2) & C.x,
        ),
    )



from __future__ import annotations

import pytest

from pydiverse.transform import C
from pydiverse.transform.core import AbstractTableImpl, Table, dtypes
from pydiverse.transform.core.dispatchers import (
    col_to_table,
    inverse_partial,
    unwrap_tables,
    verb,
    wrap_tables,
)
from pydiverse.transform.core.expressions import Column, SymbolicExpression
from pydiverse.transform.core.expressions.translator import TypedValue
from pydiverse.transform.core.util import bidict, ordered_set, sign_peeler
from pydiverse.transform.core.verbs import (
    arrange,
    collect,
    filter,
    join,
    mutate,
    rename,
    select,
)


@pytest.fixture
def tbl1():
    return Table(MockTableImpl("mock1", ["col1", "col2"]))


@pytest.fixture
def tbl2():
    return Table(MockTableImpl("mock2", ["col1", "col2", "col3"]))


class TestTable:
    def test_getattr(self, tbl1):
        assert tbl1.col1._.name == "col1"
        assert tbl1.col2._.table == tbl1._impl

        with pytest.raises(AttributeError, match="colXXX"):
            _ = tbl1.colXXX

    def test_getitem(self, tbl1):
        assert tbl1.col1._ == tbl1["col1"]._
        assert tbl1.col2._ == tbl1["col2"]._

        assert tbl1.col2._ == tbl1[tbl1.col2]._
        assert tbl1.col2._ == tbl1[C.col2]._

    def test_setitem(self, tbl1):
        tbl1["col1"] = 1
        tbl1[tbl1.col1] = 1
        tbl1[C.col1] = 1

    def test_iter(self, tbl1, tbl2):
        assert len(list(tbl1)) == len(list(tbl1._impl.selected_cols()))
        assert len(list(tbl2)) == len(list(tbl2._impl.selected_cols()))

        assert repr(list(tbl1)) == repr([tbl1.col1, tbl1.col2])
        assert repr(list(tbl2)) == repr([tbl2.col1, tbl2.col2, tbl2.col3])

        assert repr(list(tbl2 >> select(tbl2.col2))) == repr([tbl2.col2])

        joined = tbl1 >> join(tbl2 >> select(tbl2.col3), tbl1.col1 == tbl2.col2, "left")
        assert repr(list(joined)) == repr([tbl1.col1, tbl1.col2, joined.col3_mock2])

    def test_dir(self, tbl1):
        assert dir(tbl1) == ["col1", "col2"]
        assert dir(tbl1 >> mutate(x=tbl1.col1)) == ["col1", "col2", "x"]

    def test_contains(self, tbl1, tbl2):
        assert tbl1.col1 in tbl1
        assert tbl1.col2 in tbl1

        assert tbl1.col1 not in tbl2
        assert tbl1.col2 not in tbl2

        assert C.col1 in tbl1
        assert C.col2 in tbl1
        assert C.col3 not in tbl1

        assert C.col1 in tbl2
        assert C.col2 in tbl2
        assert C.col3 in tbl2
        assert C.col4 not in tbl2

        assert all(col in tbl1 for col in tbl1)
        assert all(col in tbl2 for col in tbl2)


class TestDispatchers:
    def test_inverse_partial(self):
        def x(a, b, c):
            return (a, b, c)

        assert inverse_partial(x, 1, 2)(0) == (0, 1, 2)
        assert inverse_partial(x, 1, c=2)(0, c=3) == (0, 1, 3)

    def test_pipeable(self):
        @verb
        def add(v1, v2):
            return v1 + v2

        @verb
        def subtract(v1, v2):
            return v1 - v2

        assert 1 >> add(2) == 3

        add_10 = add(5) >> add(5)
        assert 5 >> add_10 == 15

        assert 5 >> subtract(3) == 2
        assert 5 >> add_10 >> subtract(5) == 10

    def test_col_to_table(self, tbl1):
        assert col_to_table(15) == 15
        assert col_to_table(tbl1) == tbl1

        c1_tbl = col_to_table(tbl1.col1._)
        assert isinstance(c1_tbl, AbstractTableImpl)
        assert c1_tbl.available_cols == {tbl1.col1._.uuid}
        assert list(c1_tbl.named_cols.fwd) == ["col1"]

    def test_unwrap_tables(self):
        impl_1 = MockTableImpl("impl_1", dict())
        impl_2 = MockTableImpl("impl_2", dict())
        tbl_1 = Table(impl_1)
        tbl_2 = Table(impl_2)

        assert unwrap_tables(15) == 15
        assert unwrap_tables(impl_1) == impl_1
        assert unwrap_tables(tbl_1) == impl_1

        assert unwrap_tables([tbl_1]) == [impl_1]
        assert unwrap_tables([[tbl_1], tbl_2]) == [[impl_1], impl_2]

        assert unwrap_tables((tbl_1, tbl_2)) == (impl_1, impl_2)
        assert unwrap_tables((tbl_1, (tbl_2, 15))) == (impl_1, (impl_2, 15))

        assert unwrap_tables({tbl_1: tbl_1, 15: (15, tbl_2)}) == {
            tbl_1: impl_1,
            15: (15, impl_2),
        }

    def test_wrap_tables(self):
        impl_1 = MockTableImpl("impl_1", dict())
        impl_2 = MockTableImpl("impl_2", dict())
        tbl_1 = Table(impl_1)
        tbl_2 = Table(impl_2)

        assert wrap_tables(15) == 15
        assert wrap_tables(tbl_1) == tbl_1
        assert wrap_tables(impl_1) == tbl_1

        assert wrap_tables([impl_1]) == [tbl_1]
        assert wrap_tables([impl_1, [impl_2]]) == [tbl_1, [tbl_2]]

        assert wrap_tables((impl_1,)) == (tbl_1,)


class TestBuiltinVerbs:
    def test_collect(self, tbl1):
        assert (tbl1 >> collect()) == ["col1", "col2"]

    def test_select(self, tbl1, tbl2):
        # Normal Selection
        assert (tbl1 >> select() >> collect()) == []
        assert (tbl1 >> select(tbl1.col1) >> collect()) == ["col1"]

        with pytest.raises(ValueError):
            tbl1 >> select(tbl2.col1)

        with pytest.raises(TypeError):
            tbl1 >> select(tbl1.col1 + tbl1.col1)

        # Selection with Ellipsis ...
        assert (tbl1 >> select(...) >> collect()) == ["col1", "col2"]
        assert (tbl1 >> select() >> select(...) >> collect()) == ["col1", "col2"]
        assert (tbl1 >> mutate(x=C.col1) >> select(...) >> collect()) == [
            "col1",
            "col2",
            "x",
        ]
        assert (
            tbl1
            >> select()
            >> mutate(x=C.col1, col1=C.col2)
            >> select(...)
            >> select(C.col1)
            >> select(...)
            >> collect()
        ) == ["col1", "col2", "x"]

        # Negative Selection
        assert (tbl1 >> select(-C.col1) >> collect()) == ["col2"]
        assert (tbl1 >> select(C.col1) >> select(-C.col1) >> collect()) == []
        assert (tbl1 >> mutate(x=C.col1) >> select(-C.x) >> collect()) == [
            "col1",
            "col2",
        ]
        assert (tbl1 >> mutate(x=C.col1) >> select(-C.col1) >> collect()) == [
            "col2",
            "x",
        ]
        assert (
            tbl1 >> mutate(x=C.col1, y=C.col2) >> select(-C.col1, -C.y) >> collect()
        ) == ["col2", "x"]

        with pytest.raises(ValueError):
            tbl1 >> select(-C.col1, C.col1)
        with pytest.raises(ValueError):
            tbl1 >> select(-C.col1, C.col2)
        with pytest.raises(ValueError):
            tbl1 >> select(C.col1, -C.col2)
        with pytest.raises(TypeError):
            tbl1 >> select(..., -C.col2)

        assert (tbl1 >> select(--C.col1) >> collect()) == ["col1"]  # noqa: B002
        assert (tbl1 >> select(+-+-C.col1) >> collect()) == ["col1"]

    def test_rename(self, tbl2):
        def assert_rename(name_map, expected):
            assert (tbl2 >> rename(name_map) >> collect()) == expected

        assert_rename({}, ["col1", "col2", "col3"])
        assert_rename({"col1": "col1"}, ["col1", "col2", "col3"])

        assert_rename({"col1": "1st col"}, ["1st col", "col2", "col3"])
        assert_rename({"col1": "A", "col2": "B", "col3": "C"}, ["A", "B", "C"])

        # Order of name_map shouldn't matter
        assert_rename({"col3": "C", "col2": "B", "col1": "A"}, ["A", "B", "C"])

        # Swap names
        assert_rename({"col1": "col2", "col2": "col1"}, ["col2", "col1", "col3"])

        # Rename + Select
        assert tbl2 >> select(C.col1) >> rename(
            {"col1": "A", "col2": "col1"}
        ) >> collect() == ["A"]

        assert tbl2 >> select(-C.col1) >> rename(
            {"col1": "A", "col2": "col1", "col3": "col2"}
        ) >> select(*tbl2) >> collect() == ["A", "col1", "col2"]

        assert tbl2 >> rename({"col1": "col2", "col2": "col1"}) >> select(
            tbl2.col1, tbl2.col2
        ) >> collect() == ["col2", "col1"]

        # Key and value must be strings
        with pytest.raises(TypeError):
            tbl2 >> rename({1: "name"})
        with pytest.raises(TypeError):
            tbl2 >> rename({"col1": 1})

        # Try to rename cols that don't exist
        with pytest.raises(KeyError):
            tbl2 >> rename({"NONE": "SOMETHING"})
        with pytest.raises(KeyError):
            tbl2 >> rename({"name": "name"})

        # Map multiple columns to the same name
        with pytest.raises(ValueError):
            tbl2 >> rename({"col1": "A", "col2": "A"})
        with pytest.raises(ValueError):
            tbl2 >> rename({"col1": "A"}) >> rename({"col2": "A"})

        # Rename column to one that already exists
        with pytest.raises(ValueError):
            tbl2 >> rename({"col1": "col2"})
        with pytest.raises(ValueError):
            tbl2 >> rename({"col1": "A"}) >> rename({"col2": "A"})
        with pytest.raises(ValueError):
            tbl2 >> mutate(A=C.col1) >> select(-C.A) >> rename({"col1": "A"})

    def test_mutate(self, tbl1, tbl2):
        assert (tbl1 >> mutate() >> collect()) == ["col1", "col2"]
        assert (tbl1 >> mutate(x=tbl1.col1) >> collect()) == ["col1", "col2", "x"]
        assert (tbl1 >> select() >> mutate(x=tbl1.col1) >> collect()) == ["x"]

        with pytest.raises(ValueError):
            tbl1 >> mutate(x=tbl2.col1)

        t = tbl1 >> mutate(x=tbl1.col1 + tbl1.col2)
        t >> mutate(y=t.x)

    def test_join(self, tbl1, tbl2):
        assert len(tbl1 >> join(tbl2, tbl1.col1 == tbl2.col1, "left") >> collect()) == 5

        assert (
            len(
                tbl1
                >> select()
                >> join(tbl2, tbl1.col1 == tbl2.col1, "left")
                >> collect()
            )
            == 3
        )
        assert (
            len(
                tbl1
                >> join(tbl2 >> select(), tbl1.col1 == tbl2.col1, "left")
                >> collect()
            )
            == 2
        )
        assert (
            len(
                tbl1
                >> select()
                >> join(tbl2 >> select(), tbl1.col1 == tbl2.col1, "left")
                >> select(tbl1.col1, tbl2.col2)
                >> collect()
            )
            == 2
        )

        with pytest.raises(ValueError):
            # self join without alias
            tbl1 >> join(tbl1, tbl1.col1 == tbl1.col1, "inner")

        # Test that joined columns are accessible
        t = tbl1 >> join(tbl2, tbl1.col1 == tbl2.col1, "left")
        t >> select(t.col1, t.col1_mock2)

        # Select Ellipsis
        assert (
            tbl1
            >> join(tbl2 >> select(), tbl1.col1 == tbl2.col1, "left")
            >> select(...)
            >> collect()
        ) == ["col1", "col2", "col1_mock2", "col2_mock2", "col3_mock2"]

    def test_filter(self, tbl1, tbl2):
        tbl1 >> filter()
        tbl1 >> filter(tbl1.col1)

        with pytest.raises(ValueError):
            tbl1 >> filter(tbl2.col1)

    def test_arrange(self, tbl1, tbl2):
        tbl1 >> arrange(tbl1.col1)
        tbl1 >> arrange(-tbl1.col1)
        tbl1 >> arrange(tbl1.col1, tbl1.col2)
        tbl1 >> arrange(tbl1.col1, -tbl1.col2)

        with pytest.raises(ValueError):
            tbl1 >> arrange(tbl2.col1)
        with pytest.raises(ValueError):
            tbl1 >> arrange(tbl1.col1, -tbl2.col1)

    def test_col_pipeable(self, tbl1, tbl2):
        result = tbl1.col1 >> mutate(x=tbl1.col1 * 2)

        assert result._impl.selects == ordered_set(["col1", "x"])
        assert list(result._impl.named_cols.fwd) == ["col1", "x"]

        with pytest.raises(TypeError):
            (tbl1.col1 + 2) >> mutate(x=1)


class TestDataStructures:
    def test_bidict(self):
        d = bidict({"a": 1, "b": 2, "c": 3})

        assert len(d) == 3
        assert tuple(d.fwd.keys()) == ("a", "b", "c")
        assert tuple(d.fwd.values()) == (1, 2, 3)

        assert tuple(d.fwd.keys()) == tuple(d.bwd.values())
        assert tuple(d.bwd.keys()) == tuple(d.fwd.values())

        d.fwd["d"] = 4
        d.bwd[4] = "x"

        assert tuple(d.fwd.keys()) == ("a", "b", "c", "x")
        assert tuple(d.fwd.values()) == (1, 2, 3, 4)
        assert tuple(d.fwd.keys()) == tuple(d.bwd.values())
        assert tuple(d.bwd.keys()) == tuple(d.fwd.values())

        assert "x" in d.fwd
        assert "d" not in d.fwd

        d.clear()

        assert len(d) == 0
        assert len(d.fwd.items()) == len(d.fwd) == 0
        assert len(d.bwd.items()) == len(d.bwd) == 0

        with pytest.raises(ValueError):
            bidict({"a": 1, "b": 1})

    def test_ordered_set(self):
        s = ordered_set([0, 1, 2])
        assert list(s) == [0, 1, 2]

        s.add(1)  # Already in set -> Noop
        assert list(s) == [0, 1, 2]
        s.add(3)  # Not in set -> add to the end
        assert list(s) == [0, 1, 2, 3]

        s.remove(1)
        assert list(s) == [0, 2, 3]
        s.add(1)
        assert list(s) == [0, 2, 3, 1]

        assert 1 in s
        assert 4 not in s
        assert len(s) == 4

        s.clear()
        assert len(s) == 0
        assert list(s) == []

        # Set Operations

        s1 = ordered_set([0, 1, 2, 3])
        s2 = ordered_set([5, 4, 3, 2])

        assert not s1.isdisjoint(s2)
        assert list(s1 | s2) == [0, 1, 2, 3, 5, 4]
        assert list(s1 ^ s2) == [0, 1, 5, 4]
        assert list(s1 & s2) == [3, 2]
        assert list(s1 - s2) == [0, 1]

        # Pop order

        s = ordered_set([0, 1, 2, 3])
        assert s.pop() == 0
        assert s.pop() == 1
        assert s.pop_back() == 3
        assert s.pop_back() == 2


class TestUtil:
    def test_sign_peeler(self):
        x = object()
        sx = SymbolicExpression(x)
        assert sign_peeler(sx._) == (x, True)
        assert sign_peeler((+sx)._) == (x, True)
        assert sign_peeler((-sx)._) == (x, False)
        assert sign_peeler((--sx)._) == (x, True)  # noqa: B002
        assert sign_peeler((--+sx)._) == (x, True)  # noqa: B002
        assert sign_peeler((-++--sx)._) == (x, False)  # noqa: B002


class MockTableImpl(AbstractTableImpl):
    def __init__(self, name, col_names):
        super().__init__(
            name, {name: Column(name, self, dtypes.Int()) for name in col_names}
        )

    def resolve_lambda_cols(self, expr):
        return expr

    def collect(self):
        return list(self.selects)

    class ExpressionCompiler(AbstractTableImpl.ExpressionCompiler):
        def _translate(self, expr, **kwargs):
            return TypedValue(None, dtypes.Int())



from __future__ import annotations

import pytest

from pydiverse.transform.core import dtypes
from pydiverse.transform.core.registry import (
    OperatorRegistry,
    OperatorSignature,
)
from pydiverse.transform.ops import Operator


def assert_signature(
    s: OperatorSignature, args: list[dtypes.DType], rtype: dtypes.DType
):
    assert len(s.args) == len(args)

    for actual, expected in zip(s.args, args):
        assert expected.same_kind(actual)

    assert rtype.same_kind(s.rtype)


class TestOperatorSignature:
    def test_parse_simple(self):
        s = OperatorSignature.parse("int, int -> int")
        assert_signature(s, [dtypes.Int(), dtypes.Int()], dtypes.Int())

        s = OperatorSignature.parse("bool->bool ")
        assert_signature(s, [dtypes.Bool()], dtypes.Bool())

        s = OperatorSignature.parse("-> int")
        assert_signature(s, [], dtypes.Int())

        with pytest.raises(ValueError):
            OperatorSignature.parse("int, int -> ")
        with pytest.raises(ValueError):
            OperatorSignature.parse("int, int -> int, int")

        with pytest.raises(ValueError):
            OperatorSignature.parse("o#r -> int")
        with pytest.raises(ValueError):
            OperatorSignature.parse("int -> a#")

    def test_parse_template(self):
        s = OperatorSignature.parse("T, int -> int")
        assert isinstance(s.args[0], dtypes.Template)

        s = OperatorSignature.parse("T -> T")
        assert isinstance(s.args[0], dtypes.Template)
        assert isinstance(s.rtype, dtypes.Template)

        with pytest.raises(ValueError):
            OperatorSignature.parse("T, T -> U")

    def test_parse_varargs(self):
        s = OperatorSignature.parse("int, str... -> int")
        assert not s.args[0].vararg
        assert s.args[1].vararg

        s = OperatorSignature.parse("int... -> bool")
        assert s.args[0].vararg

        with pytest.raises(ValueError):
            OperatorSignature.parse("int..., str -> int")
        with pytest.raises(ValueError):
            OperatorSignature.parse("int, str -> int...")

        s0 = OperatorSignature.parse("int, str    -> int")
        s1 = OperatorSignature.parse("int, str... -> int")

        assert not s0.is_vararg
        assert s1.is_vararg

    def test_parse_const(self):
        s = OperatorSignature.parse("const int -> int")
        assert s.args[0].const
        assert not s.rtype.const


def parse_dtypes(*strings):
    return [dtypes.dtype_from_string(s) for s in strings]


class TestOperatorRegistry:
    class Op1(Operator):
        name = "op1"
        ftype = None

    class Op2(Operator):
        name = "op2"
        ftype = None

    class Op3(Operator):
        name = "op3"
        ftype = None

    def test_simple(self):
        op1 = self.Op1()
        op2 = self.Op2()

        reg = OperatorRegistry("TestRegistry")
        reg.register_op(op1)
        reg.register_op(op2)

        reg.add_implementation(op1, lambda: 1, "int, int -> int")
        reg.add_implementation(op1, lambda: 2, "str, str -> str")

        reg.add_implementation(op2, lambda: 10, "int, int -> int")
        reg.add_implementation(op2, lambda: 20, "str, str -> str")

        assert reg.get_implementation("op1", parse_dtypes("int", "int"))() == 1
        assert isinstance(
            reg.get_implementation("op1", parse_dtypes("int", "int")).rtype,
            dtypes.Int,
        )
        assert reg.get_implementation("op2", parse_dtypes("int", "int"))() == 10

        assert reg.get_implementation("op1", parse_dtypes("str", "str"))() == 2
        assert reg.get_implementation("op2", parse_dtypes("str", "str"))() == 20

        with pytest.raises(ValueError):
            reg.get_implementation("op1", parse_dtypes("int", "str"))
        with pytest.raises(ValueError):
            reg.get_implementation(
                "not_implemented",
                parse_dtypes(
                    "int",
                ),
            )

        reg.add_implementation(op1, lambda: 100, "-> int")
        assert reg.get_implementation("op1", tuple())() == 100

    def test_template(self):
        reg = OperatorRegistry("TestRegistry")

        op1 = self.Op1()
        op2 = self.Op2()
        op3 = self.Op3()

        reg.register_op(op1)
        reg.register_op(op2)
        reg.register_op(op3)

        reg.add_implementation(op1, lambda: 1, "T, T -> bool")
        reg.add_implementation(op1, lambda: 2, "T, U -> U")

        with pytest.raises(ValueError, match="already defined"):
            reg.add_implementation(op1, lambda: 3, "T, U -> U")

        assert reg.get_implementation("op1", parse_dtypes("int", "int"))() == 1
        assert reg.get_implementation("op1", parse_dtypes("int", "str"))() == 2
        # int can be promoted to float; results in "float, float -> bool" signature
        assert reg.get_implementation("op1", parse_dtypes("int", "float"))() == 1
        assert reg.get_implementation("op1", parse_dtypes("float", "int"))() == 1

        # More template matching... Also check matching precedence
        reg.add_implementation(op2, lambda: 1, "int, int, int -> int")
        reg.add_implementation(op2, lambda: 2, "int, str, T -> int")
        reg.add_implementation(op2, lambda: 3, "int, T, str -> int")
        reg.add_implementation(op2, lambda: 4, "int, T, T -> int")
        reg.add_implementation(op2, lambda: 5, "T, T, T -> int")
        reg.add_implementation(op2, lambda: 6, "A, T, T -> int")

        assert reg.get_implementation("op2", parse_dtypes("int", "int", "int"))() == 1
        assert reg.get_implementation("op2", parse_dtypes("int", "str", "str"))() == 2
        assert reg.get_implementation("op2", parse_dtypes("int", "int", "str"))() == 3
        assert reg.get_implementation("op2", parse_dtypes("int", "bool", "bool"))() == 4
        assert reg.get_implementation("op2", parse_dtypes("str", "str", "str"))() == 5
        assert reg.get_implementation("op2", parse_dtypes("float", "str", "str"))() == 6

        with pytest.raises(ValueError):
            reg.get_implementation("op2", parse_dtypes("int", "bool", "float"))

        # Return type
        reg.add_implementation(op3, lambda: 1, "T -> T")
        reg.add_implementation(op3, lambda: 2, "int, T, U -> T")
        reg.add_implementation(op3, lambda: 3, "str, T, U -> U")

        with pytest.raises(ValueError, match="already defined."):
            reg.add_implementation(op3, lambda: 4, "int, T, U -> U")

        assert isinstance(
            reg.get_implementation("op3", parse_dtypes("str")).rtype,
            dtypes.String,
        )
        assert isinstance(
            reg.get_implementation("op3", parse_dtypes("int")).rtype,
            dtypes.Int,
        )
        assert isinstance(
            reg.get_implementation("op3", parse_dtypes("int", "int", "float")).rtype,
            dtypes.Int,
        )
        assert isinstance(
            reg.get_implementation("op3", parse_dtypes("str", "int", "float")).rtype,
            dtypes.Float,
        )

    def test_vararg(self):
        reg = OperatorRegistry("TestRegistry")

        op1 = self.Op1()
        reg.register_op(op1)

        reg.add_implementation(op1, lambda: 1, "int... -> int")
        reg.add_implementation(op1, lambda: 2, "int, int... -> int")
        reg.add_implementation(op1, lambda: 3, "int, T... -> T")

        assert (
            reg.get_implementation(
                "op1",
                parse_dtypes(
                    "int",
                ),
            )()
            == 1
        )
        assert reg.get_implementation("op1", parse_dtypes("int", "int"))() == 2
        assert reg.get_implementation("op1", parse_dtypes("int", "int", "int"))() == 2
        assert reg.get_implementation("op1", parse_dtypes("int", "str", "str"))() == 3

        assert isinstance(
            reg.get_implementation("op1", parse_dtypes("int", "str", "str")).rtype,
            dtypes.String,
        )

    def test_variant(self):
        op1 = self.Op1()

        reg = OperatorRegistry("TestRegistry")
        reg.register_op(op1)

        with pytest.raises(ValueError):
            reg.add_implementation(op1, lambda: 2, "-> int", variant="VAR")

        reg.add_implementation(op1, lambda: 1, "-> int")
        reg.add_implementation(op1, lambda: 2, "-> int", variant="VAR")

        assert reg.get_implementation("op1", tuple())() == 1
        assert reg.get_implementation("op1", tuple()).get_variant("VAR")() == 2

        with pytest.raises(ValueError):
            reg.add_implementation(op1, lambda: 2, "-> int", variant="VAR")



from __future__ import annotations

from pydiverse.transform import C
from pydiverse.transform.core.verbs import (
    mutate,
    select,
)
from tests.util import assert_result_equal


def test_simple_select(df1):
    assert_result_equal(df1, lambda t: t >> select(t.col1))
    assert_result_equal(df1, lambda t: t >> select(t.col2))


def test_reorder(df1):
    assert_result_equal(df1, lambda t: t >> select(t.col2, t.col1))


def test_ellipsis(df3):
    assert_result_equal(df3, lambda t: t >> select(...))
    assert_result_equal(df3, lambda t: t >> select(t.col1) >> select(...))
    assert_result_equal(
        df3, lambda t: t >> mutate(x=t.col1 * 2) >> select() >> select(...)
    )


def test_negative_select(df3):
    assert_result_equal(df3, lambda t: t >> select(-t.col1))
    assert_result_equal(df3, lambda t: t >> select(-C.col1, -t.col2))
    assert_result_equal(
        df3,
        lambda t: t >> select() >> mutate(x=t.col1 * 2) >> select(-C.col3),
    )



from __future__ import annotations

from pydiverse.transform import C
from pydiverse.transform import functions as f
from pydiverse.transform.core.verbs import (
    group_by,
    mutate,
    summarise,
)
from pydiverse.transform.errors import ExpressionTypeError, FunctionTypeError
from tests.util import assert_result_equal


def test_mutate_case_ewise(df4):
    assert_result_equal(
        df4,
        lambda t: t
        >> mutate(
            x=C.col1.case(
                (0, 1),
                (1, 2),
                (2, 2),
            ),
            y=C.col1.case(
                (0, 0),
                (1, None),
                default=10.5,
            ),
        ),
    )

    assert_result_equal(
        df4,
        lambda t: t
        >> mutate(
            x=f.case(
                (C.col1 == C.col2, 1),
                (C.col2 == C.col3, 2),
                default=(C.col1 + C.col2),
            )
        ),
    )


def test_mutate_case_window(df4):
    assert_result_equal(
        df4,
        lambda t: t
        >> mutate(
            x=f.case(
                (C.col1.max() == 1, 1),
                (C.col1.max() == 2, 2),
                (C.col1.max() == 3, 3),
                (C.col1.max() == 4, 4),
            )
        ),
    )

    assert_result_equal(
        df4,
        lambda t: t
        >> mutate(
            u=C.col1.shift(1, 1729, arrange=[-t.col3, t.col4]),
            x=C.col1.shift(1, 0, arrange=[C.col4]).case(
                (1, C.col2.shift(1, -1, arrange=[C.col2, C.col4])),
                (2, C.col3.shift(2, -2, arrange=[C.col3, C.col4])),
            ),
        ),
    )

    # Can't nest window in window
    assert_result_equal(
        df4,
        lambda t: t
        >> mutate(
            x=C.col1.shift(1, 0, arrange=[C.col4])
            .case(
                (1, 2),
                (2, 3),
            )
            .shift(1, -1, arrange=[-C.col4])
        ),
        may_throw=True,
    )


def test_summarise_case(df4):
    assert_result_equal(
        df4,
        lambda t: t
        >> group_by(
            C.col1,
        )
        >> summarise(
            x=C.col2.max().case(
                (0, C.col1.min()),  # Int
                (1, C.col2.mean() + 0.5),  # Float
                (2, 2),  # ftype=EWISE
            ),
            y=f.case(
                (C.col2.max() > 2, 1),
                (C.col2.max() < 2, C.col2.min()),
                default=C.col3.mean(),
            ),
        ),
    )


def test_invalid_value_dtype(df4):
    # Incompatible types String and Float
    assert_result_equal(
        df4,
        lambda t: t
        >> mutate(
            x=C.col1.case(
                (0, "a"),
                (1, 1.1),
            )
        ),
        exception=ExpressionTypeError,
    )


def test_invalid_result_dtype(df4):
    # Invalid result type: none
    assert_result_equal(
        df4,
        lambda t: t
        >> mutate(
            x=f.case(
                default=None,
            )
        ),
        exception=ExpressionTypeError,
    )


def test_invalid_ftype(df1):
    assert_result_equal(
        df1,
        lambda t: t
        >> summarise(
            x=f.rank(arrange=[C.col1]).case(
                (1, C.col1.max()),
                default=None,
            )
        ),
        exception=FunctionTypeError,
    )

    assert_result_equal(
        df1,
        lambda t: t
        >> summarise(
            x=f.case(
                (f.rank(arrange=[C.col1]) == 1, 1),
                default=None,
            )
        ),
        exception=FunctionTypeError,
    )



from __future__ import annotations

from pydiverse.transform import C
from pydiverse.transform.core import functions as f
from pydiverse.transform.core.verbs import (
    arrange,
    filter,
    group_by,
    left_join,
    mutate,
    select,
    slice_head,
    summarise,
)
from tests.util import assert_result_equal, full_sort


def test_simple(df3):
    assert_result_equal(df3, lambda t: t >> arrange(*t) >> slice_head(1))
    assert_result_equal(df3, lambda t: t >> arrange(*t) >> slice_head(10))
    assert_result_equal(df3, lambda t: t >> arrange(*t) >> slice_head(100))

    assert_result_equal(df3, lambda t: t >> arrange(*t) >> slice_head(1, offset=8))
    assert_result_equal(df3, lambda t: t >> arrange(*t) >> slice_head(10, offset=8))
    assert_result_equal(df3, lambda t: t >> arrange(*t) >> slice_head(100, offset=8))

    assert_result_equal(df3, lambda t: t >> arrange(*t) >> slice_head(1, offset=100))
    assert_result_equal(df3, lambda t: t >> arrange(*t) >> slice_head(10, offset=100))
    assert_result_equal(df3, lambda t: t >> arrange(*t) >> slice_head(100, offset=100))


def test_chained(df3):
    assert_result_equal(
        df3,
        lambda t: t >> arrange(*t) >> slice_head(1) >> arrange(*t) >> slice_head(1),
    )
    assert_result_equal(
        df3,
        lambda t: t >> arrange(*t) >> slice_head(10) >> arrange(*t) >> slice_head(5),
    )
    assert_result_equal(
        df3,
        lambda t: t >> arrange(*t) >> slice_head(100) >> arrange(*t) >> slice_head(5),
    )

    assert_result_equal(
        df3,
        lambda t: t
        >> arrange(*t)
        >> slice_head(2, offset=5)
        >> arrange(*t)
        >> slice_head(2, offset=1),
    )
    assert_result_equal(
        df3,
        lambda t: t
        >> arrange(*t)
        >> slice_head(10, offset=8)
        >> arrange(*t)
        >> slice_head(10, offset=1),
    )


def test_with_select(df3):
    assert_result_equal(
        df3,
        lambda t: t >> select() >> arrange(*t) >> slice_head(4, offset=2) >> select(*t),
    )


def test_with_mutate(df3):
    assert_result_equal(
        df3,
        lambda t: t
        >> mutate(a=C.col1 * 2)
        >> arrange(*t)
        >> slice_head(4, offset=2)
        >> mutate(b=C.col2 + C.a),
    )


def test_with_join(df1, df2):
    assert_result_equal(
        (df1, df2),
        lambda t, u: t
        >> full_sort()
        >> arrange(*t)
        >> slice_head(3)
        >> left_join(u, t.col1 == u.col1)
        >> full_sort(),
    )

    assert_result_equal(
        (df1, df2),
        lambda t, u: t
        >> left_join(u >> arrange(*t) >> slice_head(2, offset=1), t.col1 == u.col1)
        >> full_sort(),
        exception=ValueError,
        may_throw=True,
    )


def test_with_filter(df3):
    assert_result_equal(
        df3,
        lambda t: t
        >> filter(t.col4 % 2 == 0)
        >> arrange(*t)
        >> slice_head(4, offset=2),
    )

    assert_result_equal(
        df3,
        lambda t: t >> arrange(*t) >> slice_head(4, offset=2) >> filter(t.col1 == 1),
    )

    assert_result_equal(
        df3,
        lambda t: t
        >> filter(t.col4 % 2 == 0)
        >> arrange(*t)
        >> slice_head(4, offset=2)
        >> filter(t.col1 == 1),
    )


def test_with_arrange(df3):
    assert_result_equal(
        df3,
        lambda t: t
        >> mutate(x=t.col4 - (t.col1 * t.col2))
        >> arrange(C.x, *t)
        >> slice_head(4, offset=2),
    )

    assert_result_equal(
        df3,
        lambda t: t
        >> mutate(x=(t.col1 * t.col2))
        >> arrange(*t)
        >> slice_head(4)
        >> arrange(-C.x, C.col5),
    )


def test_with_group_by(df3):
    assert_result_equal(
        df3,
        lambda t: t
        >> arrange(*t)
        >> slice_head(1)
        >> group_by(C.col1)
        >> mutate(x=f.count()),
    )

    assert_result_equal(
        df3,
        lambda t: t
        >> arrange(C.col1, *t)
        >> slice_head(6, offset=1)
        >> group_by(C.col1)
        >> select()
        >> mutate(x=C.col4.mean()),
    )

    assert_result_equal(
        df3,
        lambda t: t
        >> mutate(key=C.col4 % (C.col3 + 1))
        >> arrange(C.key, *t)
        >> slice_head(4)
        >> group_by(C.key)
        >> summarise(x=f.count()),
    )


def test_with_summarise(df3):
    assert_result_equal(
        df3,
        lambda t: t >> arrange(*t) >> slice_head(4) >> summarise(count=f.count()),
    )

    assert_result_equal(
        df3,
        lambda t: t >> arrange(*t) >> slice_head(4) >> summarise(c3_mean=C.col3.mean()),
    )



from __future__ import annotations

from pydiverse.transform.ops.core import ElementWise, OperatorExtension, Unary
from pydiverse.transform.ops.logical import Logical
from pydiverse.transform.ops.numeric import Add, RAdd

__all__ = [
    "StrAdd",
    "StrRAdd",
    "StrStrip",
    "StrLen",
    "StrToUpper",
    "StrToLower",
    "StrReplaceAll",
    "StrStartsWith",
    "StrEndsWith",
    "StrContains",
    "StrSlice",
]


class StrAdd(OperatorExtension):
    operator = Add
    signatures = [
        "str, str -> str",
    ]


class StrRAdd(OperatorExtension):
    operator = RAdd
    signatures = [
        "str, str -> str",
    ]


####


class StrUnary(ElementWise, Unary):
    signatures = [
        "str -> str",
    ]


class StrStrip(StrUnary):
    name = "str.strip"


class StrLen(StrUnary):
    name = "str.len"
    signatures = [
        "str -> int",
    ]


class StrToUpper(StrUnary):
    name = "str.to_upper"


class StrToLower(StrUnary):
    name = "str.to_lower"


class StrReplaceAll(ElementWise):
    name = "str.replace_all"
    signatures = [
        "str, const str, const str -> str",
    ]


class StrStartsWith(ElementWise, Logical):
    name = "str.starts_with"
    signatures = [
        "str, const str -> bool",
    ]


class StrEndsWith(ElementWise, Logical):
    name = "str.ends_with"
    signatures = [
        "str, const str -> bool",
    ]


class StrContains(ElementWise, Logical):
    name = "str.contains"
    signatures = [
        "str, const str -> bool",
    ]


class StrSlice(ElementWise):
    name = "str.slice"
    signatures = ["str, int, int -> str"]



from __future__ import annotations

from pydiverse.transform import Table, verb
from pydiverse.transform.core.verbs import arrange


@verb
def full_sort(t: Table):
    """
    Ordering after join is not determined.
    This helper applies a deterministic ordering to a table.
    """
    return t >> arrange(*t)



from __future__ import annotations

from pydiverse.transform import C
from pydiverse.transform.core.verbs import (
    mutate,
    select,
)
from pydiverse.transform.errors import ExpressionTypeError
from tests.util import assert_result_equal


def test_noop(df2):
    assert_result_equal(
        df2, lambda t: t >> mutate(col1=t.col1, col2=t.col2, col3=t.col3)
    )


def test_multiply(df1):
    assert_result_equal(df1, lambda t: t >> mutate(x2=t.col1 * 2))
    assert_result_equal(df1, lambda t: t >> select() >> mutate(x2=t.col1 * 2))


def test_reorder(df2):
    assert_result_equal(df2, lambda t: t >> mutate(col1=t.col2, col2=t.col1))

    assert_result_equal(
        df2,
        lambda t: t
        >> mutate(col1=t.col2, col2=t.col1)
        >> mutate(col1=t.col2, col2=C.col3, col3=C.col2),
    )


def test_literals(df1):
    assert_result_equal(df1, lambda t: t >> mutate(x=1))
    assert_result_equal(df1, lambda t: t >> mutate(x=1.1))
    assert_result_equal(df1, lambda t: t >> mutate(x=True))
    assert_result_equal(df1, lambda t: t >> mutate(x="test"))


def test_none(df4):
    assert_result_equal(
        df4, lambda t: t >> mutate(x=None), exception=ExpressionTypeError
    )
    assert_result_equal(
        df4,
        lambda t: t
        >> mutate(
            x1=t.col1.is_null(),
            y1=~t.col2.is_null(),
        ),
    )


def test_mutate_bool_expr(df4):
    assert_result_equal(
        df4,
        lambda t: t
        >> mutate(x=t.col1 <= t.col2, y=(t.col3 * 4) >= C.col4)
        >> mutate(xAndY=C.x & C.y),
    )
